import copy
import json
import logging
import math
import os
import time
import xml.etree.ElementTree as ET
from typing import Union

import gevent
import numpy
from pydantic import ValidationError

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import sample_centring
from mxcubecore.HardwareObjects.GenericDiffractometer import GonioHeadConfiguration
from mxcubecore.model import queue_model_objects as qmo
from mxcubecore.TaskUtils import task


class MiniDiff(HardwareObject):
    MANUAL3CLICK_MODE = "Manual 3-click"
    CENTRING_METHOD_MANUAL = MANUAL3CLICK_MODE
    C3D_MODE = "Computer automatic"
    # MOVE_TO_BEAM_MODE = "Move to Beam"

    def __init__(self, *args):
        super().__init__(*args)

        qmo.CentredPosition.set_diffractometer_motor_names(
            "phi",
            "focus",
            "phiz",
            "phiy",
            "zoom",
            "sampx",
            "sampy",
            "kappa",
            "kappa_phi",
        )

        self.phiMotor = None
        self.kappaMotor = None
        self.kappaPhiMotor = None
        self.phizMotor = None
        self.phiyMotor = None
        self.lightMotor = None
        self.zoomMotor = None
        self.sampleXMotor = None
        self.sampleYMotor = None
        self.lightWago = None
        self.currentSampleInfo = None
        self.aperture = None
        self.cryostream = None
        self.beamstop = None
        self.capillary = None

        self.pixelsPerMmY = None
        self.pixelsPerMmZ = None
        self.centredTime = 0
        self.user_confirms_centring = True
        self.do_centring = True
        self.chiAngle = 0.0

        self.connect(self, "equipmentReady", self.equipmentReady)
        self.connect(self, "equipmentNotReady", self.equipmentNotReady)

    def if_role_set_attr(self, role_name):
        obj = self.get_object_by_role(role_name)

        if obj is not None:
            setattr(self, role_name, obj)

    def init(self):
        self.centringMethods = {
            MiniDiff.MANUAL3CLICK_MODE: self.start_manual_centring,
            MiniDiff.C3D_MODE: self.start_auto_centring,
        }

        self._run_script = self.add_command(
            {
                "type": "exporter",
                "exporter_address": self.exporter_addr,
                "name": "runScript",
            },
            "runScript",
        )

        sample_centring.NUM_CENTRING_ROUNDS = self.get_property(
            "num_centering_rounds", 1
        )

        self.cancel_centring_methods = {}

        self.current_centring_procedure = None
        self.currentCentringMethod = None

        self.centringStatus = {"valid": False}

        self.chiAngle = self.get_property("chi", 0)

        try:
            phiz_ref = self["centringReferencePosition"].get_property("phiz")
        except:
            phiz_ref = None

        try:
            phiy_ref = self["centringReferencePosition"].get_property("phiy")
        except:
            phiy_ref = None

        self.phiMotor = self.get_object_by_role("phi")
        self.phizMotor = self.get_object_by_role("phiz")
        self.phiyMotor = self.get_object_by_role("phiy")
        self.zoomMotor = self.get_object_by_role("zoom")
        self.lightMotor = self.get_object_by_role("light")
        self.focusMotor = self.get_object_by_role("focus")
        self.sampleXMotor = self.get_object_by_role("sampx")
        self.sampleYMotor = self.get_object_by_role("sampy")
        self.kappaMotor = self.get_object_by_role("kappa")
        self.kappaPhiMotor = self.get_object_by_role("kappa_phi")

        self.centringPhi = sample_centring.CentringMotor(self.phiMotor, direction=-1)
        self.centringPhiz = sample_centring.CentringMotor(
            self.phizMotor, reference_position=phiz_ref
        )
        self.centringPhiy = sample_centring.CentringMotor(
            self.phiyMotor, reference_position=phiy_ref
        )
        self.centringSamplex = sample_centring.CentringMotor(self.sampleXMotor)
        self.centringSampley = sample_centring.CentringMotor(self.sampleYMotor)

        roles_to_add = ["aperture", "beamstop", "cryostream", "capillary"]

        for role in roles_to_add:
            self.if_role_set_attr(role)

        hwr = HWR.get_hardware_repository()
        wl_prop = self.get_property("wagolight")
        if wl_prop is not None:
            try:
                self.lightWago = hwr.get_hardware_object(wl_prop)
            except Exception:
                self.log.exception("")

        if self.phiMotor is not None:
            self.connect(self.phiMotor, "stateChanged", self.phiMotorStateChanged)
            self.connect(self.phiMotor, "valueChanged", self.emit_diffractometer_moved)
        else:
            self.log.error(
                "MiniDiff: phi motor is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.phizMotor is not None:
            self.connect(self.phizMotor, "stateChanged", self.phizMotorStateChanged)
            self.connect(self.phizMotor, "valueChanged", self.phizMotorMoved)
            self.connect(self.phizMotor, "valueChanged", self.emit_diffractometer_moved)
        else:
            self.log.error(
                "MiniDiff: phiz motor is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.phiyMotor is not None:
            self.connect(self.phiyMotor, "stateChanged", self.phiyMotorStateChanged)
            self.connect(self.phiyMotor, "valueChanged", self.phiyMotorMoved)
            self.connect(self.phiyMotor, "valueChanged", self.emit_diffractometer_moved)
        else:
            self.log.error(
                "MiniDiff: phiy motor is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.zoomMotor is not None:
            self.connect(
                self.zoomMotor, "valueChanged", self.zoomMotorPredefinedPositionChanged
            )

            self.connect(
                self.zoomMotor,
                "predefinedPositionChanged",
                self.zoomMotorPredefinedPositionChanged,
            )
            self.connect(self.zoomMotor, "stateChanged", self.zoomMotorStateChanged)
        else:
            self.log.error(
                "MiniDiff: zoom motor is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.sampleXMotor is not None:
            self.connect(
                self.sampleXMotor, "stateChanged", self.sampleXMotorStateChanged
            )
            self.connect(self.sampleXMotor, "valueChanged", self.sampleXMotorMoved)
            self.connect(
                self.sampleXMotor, "valueChanged", self.emit_diffractometer_moved
            )
        else:
            self.log.error(
                "MiniDiff: sampx motor is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.sampleYMotor is not None:
            self.connect(
                self.sampleYMotor, "stateChanged", self.sampleYMotorStateChanged
            )
            self.connect(self.sampleYMotor, "valueChanged", self.sampleYMotorMoved)
            self.connect(
                self.sampleYMotor, "valueChanged", self.emit_diffractometer_moved
            )
        else:
            self.log.error(
                "MiniDiff: sampx motor is not defined in minidiff equipment %s",
                str(self.name),
            )

        if HWR.beamline.sample_changer is None:
            self.log.warning(
                "MiniDiff: sample changer is not defined in minidiff equipment %s",
                str(self.name),
            )
        else:
            try:
                self.connect(
                    HWR.beamline.sample_changer,
                    "sampleIsLoaded",
                    self.sampleChangerSampleIsLoaded,
                )
            except Exception:
                self.log.exception(
                    "MiniDiff: could not connect to sample changer smart magnet"
                )
        if self.lightWago is not None:
            self.connect(self.lightWago, "wagoStateChanged", self.wagoLightStateChanged)
        else:
            self.log.warning(
                "MiniDiff: wago light is not defined in minidiff equipment %s",
                str(self.name),
            )
        if self.aperture is not None:
            self.connect(
                self.aperture, "predefinedPositionChanged", self.apertureChanged
            )
            self.connect(self.aperture, "positionReached", self.apertureChanged)

        # Agree on a correct method name, inconsistent arguments for move_to_beam, disabled temporarily
        # self.move_to_coord = self.move_to_beam()

    def set_rotation_axis_position(self, value: float):
        self._set_rotation_axis_position(value, motor_name="phiz")

    def _set_rotation_axis_position(self, value: float, motor_name="phiz"):
        self.log.info(f"Setting rotation axis ({motor_name}) position to {value}")

        fname = self.get_xml_path()
        try:
            self.log.info(f"Updating {fname}")

            tree = ET.parse(fname)
            motor_tag = (
                tree.getroot()
                .findall("centringReferencePosition")[0]
                .findall(motor_name)
            )
            motor_tag.text = str(value)
            tree.write(fname)
        except:
            self.log.info(f"Could not update {fname}")
            # raise
        else:
            self.log.info(f"Wrote {fname}")

        if motor_name == "phiz":
            self.centringPhiz = sample_centring.CentringMotor(
                self.phizMotor, reference_position=value
            )
        elif motor_name == "phiy":
            self.centringPhiy = sample_centring.CentringMotor(
                self.phiyMotor, reference_position=value
            )

        script_name = (
            "Change_AlignmentZ" if motor_name == "phiz" else "Change_AlignmentY"
        )

        try:
            self.log.info(f"Setting MD Alignment reference position")
            print(f" script name {script_name} value {value}")
            self.run_script(f"{script_name}, {value}")
        except:
            self.log.exception(f"Setting MD Alignment reference position failed")
            raise

    # Contained Objects
    # NBNB Temp[orary hack - should be cleaned up together with configuration
    @property
    def omega(self):
        """omega motor object

        Returns:
            AbstractActuator
        """
        return self.phiMotor

    @property
    def kappa(self):
        """kappa motor object

        Returns:
            AbstractActuator
        """
        return self.kappaMotor

    @property
    def kappa_phi(self):
        """kappa_phi motor object

        Returns:
            AbstractActuator
        """
        return self.kappaPhiMotor

    @property
    def centring_x(self):
        """centring_x motor object

        Returns:
            AbstractActuator
        """
        return self.sampleXMotor

    @property
    def centring_y(self):
        """centring_y motor object

        Returns:
            AbstractActuator
        """
        return self.sampleYMotor

    @property
    def alignment_x(self):
        """alignment_x motor object (also used as graphics.focus)

        Returns:
            AbstractActuator
        """
        return self.focusMotor

    @property
    def alignment_y(self):
        """alignment_y motor object

        Returns:
            AbstractActuator
        """
        return self.phiyMotor

    @property
    def alignment_z(self):
        """alignment_z motor object

        Returns:
            AbstractActuator
        """
        return self.phizMotor

    @property
    def zoom(self):
        """zoom motor object

        NBNB HACK TODO - ocnfigure this in graphics object
        (which now calls this property)

        Returns:
            AbstractActuator
        """
        return self.zoomMotor

    def set_sample_info(self, sample_info):
        self.currentSampleInfo = sample_info

    def emit_diffractometer_moved(self, *args):
        self.emit("diffractometerMoved", ())

    def is_ready(self):
        return self.get_state() and self.get_state().name == "READY"

    def is_valid(self):
        return (
            self.sampleXMotor is not None
            and self.sampleYMotor is not None
            and self.zoomMotor is not None
            and self.phiMotor is not None
            and self.phizMotor is not None
            and self.phiyMotor is not None
            and HWR.beamline.sample_view.camera is not None
        )

    def in_plate_mode(self):
        return False

    def apertureChanged(self, *args):
        # will trigger minidiffReady signal for update of beam size in video
        self.equipmentReady()

    def equipmentReady(self):
        self.emit("minidiffReady", ())

    def equipmentNotReady(self):
        self.emit("minidiffNotReady", ())

    def wagoLightStateChanged(self, state):
        pass

    def phiMotorStateChanged(self, state):
        self.emit("phiMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def phizMotorStateChanged(self, state):
        self.emit("phizMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def phiyMotorStateChanged(self, state):
        self.emit("phiyMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def getCalibrationData(self, offset):
        if self.zoomMotor is not None:
            for position in self.zoomMotor.get_property("positions", ()):
                if abs(position["offset"] - offset) <= self.zoomMotor.delta:
                    calibrationData = position["calibrationData"]
                    return (
                        float(calibrationData.pixelsPerMmY) or 0,
                        float(calibrationData.pixelsPerMmZ) or 0,
                    )
        return (None, None)

    def get_pixels_per_mm(self):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(None)
        return self.pixelsPerMmY, self.pixelsPerMmZ

    def getBeamInfo(self, callback=None):
        beam_info = HWR.beamline.beam.get_beam_info()
        if callable(callback):
            callback(beam_info)
        return beam_info

    def zoomMotorPredefinedPositionChanged(self, positionName, offset=None):
        if not positionName:
            return
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(offset)
        self.emit("zoomMotorPredefinedPositionChanged", (positionName, offset))

    def zoomMotorStateChanged(self, state):
        HWR.beamline.beam.re_emit_values()
        self.emit("zoomMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def sampleXMotorStateChanged(self, state):
        self.emit("sampxMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def sampleYMotorStateChanged(self, state):
        self.emit("sampyMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def invalidateCentring(self):
        if self.current_centring_procedure is None and self.centringStatus["valid"]:
            self.centringStatus = {"valid": False}
            self.emitProgressMessage("")
            self.emit("centringInvalid", ())

    def phizMotorMoved(self, pos):
        if time.time() - self.centredTime > 1.0:
            self.invalidateCentring()

    def phiyMotorMoved(self, pos):
        if time.time() - self.centredTime > 1.0:
            self.invalidateCentring()

    def sampleXMotorMoved(self, pos):
        if time.time() - self.centredTime > 1.0:
            self.invalidateCentring()

    def sampleYMotorMoved(self, pos):
        if time.time() - self.centredTime > 1.0:
            self.invalidateCentring()

    def sampleChangerSampleIsLoaded(self, state):
        if time.time() - self.centredTime > 1.0:
            self.invalidateCentring()

    # def getBeamPosX(self):
    #     return self.imgWidth / 2
    #
    # def getBeamPosY(self):
    #     return self.imgHeight / 2

    def move_to_beam(self, x, y):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )

        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0
        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        dx = (x - beam_pos_x) / self.pixelsPerMmY
        dy = (y - beam_pos_y) / self.pixelsPerMmZ

        phi_angle = math.radians(
            self.centringPhi.direction * self.centringPhi.get_value()
        )

        sampx = -self.centringSamplex.direction * self.centringSamplex.get_value()
        sampy = self.centringSampley.direction * self.centringSampley.get_value()

        phiy = self.centringPhiy.direction * self.centringPhiy.get_value()
        phiz = self.centringPhiz.direction * self.centringPhiz.get_value()

        rotMatrix = numpy.matrix(
            [
                [math.cos(phi_angle), -math.sin(phi_angle)],
                [math.sin(phi_angle), math.cos(phi_angle)],
            ]
        )
        invRotMatrix = numpy.array(rotMatrix.I)

        dsampx, dsampy = numpy.dot(numpy.array([0, dy]), invRotMatrix)

        chi_angle = math.radians(self.chiAngle)
        chiRot = numpy.matrix(
            [
                [math.cos(chi_angle), -math.sin(chi_angle)],
                [math.sin(chi_angle), math.cos(chi_angle)],
            ]
        )

        sx, sy = numpy.dot(numpy.array([dsampx, dsampy]), numpy.array(chiRot))

        sampx = sampx + sx
        sampy = sampy + sy
        phiy = phiy + dx

        try:
            self.centringSamplex.set_value(-sampx)
            self.centringSampley.set_value(sampy)
            self.centringPhiy.set_value(-phiy)
        except Exception:
            self.log.exception("MiniDiff: could not center to beam, aborting")

    def getAvailableCentringMethods(self):
        return self.centringMethods.keys()

    def run_custom_centring_script(self, method, sample_info):
        try:
            if method == "Manual 3-click":
                self.wait_ready(30)
                fun = self.centringMethods[method]
            else:
                time.sleep(0.5)
                self.wait_ready(60)
                self.log.info("Using MD script for sample centring")
                self.run_script("sample_centering")
                time.sleep(0.5)
                self.wait_ready(120)

                # if the centering fails move to the next sample
                try:
                    res_centering = self.get_last_task_info()
                    if (
                        res_centering[0].endswith("sample_centering.java")
                        and res_centering[6] == "-1"
                    ):
                        self.log.exception("MiniDiff: problem while centring")
                        self.emitCentringFailed()
                    else:
                        self.log.info(
                            "MiniDiff: centring went fine with %s" % str(res_centering)
                        )
                except:
                    self.log.exception("MD script for sample centering had a problem")

        except KeyError as diag:
            self.log.error("MiniDiff: unknown centring method (%s)" % str(diag))
            self.emitCentringFailed()
        else:
            try:
                if method == "Manual 3-click":
                    fun(sample_info)
                else:
                    pass

            except Exception:
                self.log.exception("MiniDiff: problem while centring")
                self.emitCentringFailed()

    def run_standard_centring(self, method, sample_info):
        self.emitCentringStarted(method)

        try:
            self.wait_ready(30)
            fun = self.centringMethods[method]
        except KeyError as diag:
            self.log.error("MiniDiff: unknown centring method (%s)" % str(diag))
            self.emitCentringFailed()
        else:
            try:
                fun(sample_info)
            except Exception:
                self.log.exception("MiniDiff: problem while centring")
                self.emitCentringFailed()

    def start_centring_method(self, method, sample_info=None):
        if not self.do_centring:
            self.emitCentringStarted(method)

            def fake_centring_procedure():
                return {"motors": {}, "method": method, "valid": True}

            self.current_centring_procedure = gevent.spawn(fake_centring_procedure)
            self.emitCentringSuccessful()
            return

        if self.current_centring_procedure is not None:
            self.log.error(
                "MiniDiff: already in centring method %s" % self.currentCentringMethod
            )
            return

        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centringStatus = {"valid": False, "startTime": curr_time}

        _use_custom = self.get_property("use_custom_centring_script", False)

        if _use_custom:
            self.run_custom_centring_script(method, sample_info)
        else:
            self.run_standard_centring(method, sample_info)

    def cancel_centring_method(self, reject=False):
        if self.current_centring_procedure is not None:
            try:
                self.current_centring_procedure.kill(block=True)
            except Exception:
                self.log.exception("MiniDiff: problem aborting the centring method")

            self.log.exception("MiniDiff: Centring canceled")

            try:
                fun = self.cancel_centring_methods[self.currentCentringMethod]
            except KeyError as diag:
                self.emitCentringFailed()
            else:
                try:
                    fun()
                except Exception:
                    self.emitCentringFailed()
        else:
            self.emitCentringFailed()

        self.emitProgressMessage("")

        if reject:
            self.reject_centring()

        self.wait_ready(30)

    def currentCentringMethod(self):
        return self.currentCentringMethod

    def save_current_position(self):
        self.centringStatus["motors"] = self.get_positions()
        self.accept_centring()

    def start_manual_centring(self, sample_info=None):
        self.log.info("Starting centring procedure ...")

        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

        self.wait_ready(5)

        self.current_centring_procedure = sample_centring.start(
            {
                "phi": self.centringPhi,
                "phiy": self.centringPhiy,
                "sampx": self.centringSamplex,
                "sampy": self.centringSampley,
                "phiz": self.centringPhiz,
            },
            self.pixelsPerMmY,
            self.pixelsPerMmZ,
            beam_pos_x,
            beam_pos_y,
            chi_angle=self.chiAngle,
        )

        self.current_centring_procedure.link(self.manualCentringDone)

    def motor_positions_to_screen(self, centred_positions_dict):
        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )
        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0
        phi_angle = math.radians(
            self.centringPhi.direction * self.centringPhi.get_value()
        )
        sampx = self.centringSamplex.direction * (
            centred_positions_dict["sampx"] - self.centringSamplex.get_value()
        )
        sampy = self.centringSampley.direction * (
            centred_positions_dict["sampy"] - self.centringSampley.get_value()
        )
        phiy = self.centringPhiy.direction * (
            centred_positions_dict["phiy"] - self.centringPhiy.get_value()
        )
        phiz = self.centringPhiz.direction * (
            centred_positions_dict["phiz"] - self.centringPhiz.get_value()
        )
        rotMatrix = numpy.matrix(
            [
                math.cos(phi_angle),
                -math.sin(phi_angle),
                math.sin(phi_angle),
                math.cos(phi_angle),
            ]
        )
        rotMatrix.shape = (2, 2)
        invRotMatrix = numpy.array(rotMatrix.I)
        dsx, dsy = (
            numpy.dot(numpy.array([sampx, sampy]), invRotMatrix) * self.pixelsPerMmY
        )
        chi_angle = math.radians(self.chiAngle)
        chiRot = numpy.matrix(
            [
                math.cos(chi_angle),
                -math.sin(chi_angle),
                math.sin(chi_angle),
                math.cos(chi_angle),
            ]
        )
        chiRot.shape = (2, 2)
        sx, sy = numpy.dot(numpy.array([0, dsy]), numpy.array(chiRot))  # .I))
        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

        x = sx + (phiy * self.pixelsPerMmY) + beam_pos_x
        y = sy + (phiz * self.pixelsPerMmZ) + beam_pos_y

        return float(x), float(y)

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()
        dx = (x - beam_pos_x) / self.pixelsPerMmY
        dy = (y - beam_pos_y) / self.pixelsPerMmZ

        self.pixelsPerMmY, self.pixelsPerMmZ = self.getCalibrationData(
            self.zoomMotor.get_value()
        )

        if None in (self.pixelsPerMmY, self.pixelsPerMmZ):
            return 0, 0

        phi_angle = math.radians(self.centringPhi.get_value())
        sampx = self.centringSamplex.get_value()
        sampy = self.centringSampley.get_value()
        phiy = self.centringPhiy.get_value()
        phiz = self.centringPhiz.get_value()

        rotMatrix = numpy.matrix(
            [
                math.cos(phi_angle),
                -math.sin(phi_angle),
                math.sin(phi_angle),
                math.cos(phi_angle),
            ]
        )
        rotMatrix.shape = (2, 2)
        invRotMatrix = numpy.array(rotMatrix.I)

        dsampx, dsampy = numpy.dot(numpy.array([0, dy]), invRotMatrix)
        sampx = sampx + dsampx
        sampy = sampy + dsampy
        phiy = phiy - dx

        #        chi_angle = math.radians(self.chiAngle)
        #        chiRot = numpy.matrix([math.cos(chi_angle), -math.sin(chi_angle),
        #                               math.sin(chi_angle), math.cos(chi_angle)])
        #        chiRot.shape = (2,2)
        #        sx, sy = numpy.dot(numpy.array([0, dsy]),
        #                           numpy.array(chiRot)) ))

        return {
            "phi": self.centringPhi.get_value(),
            "phiz": float(phiz),
            "phiy": float(phiy),
            "sampx": float(sampx),
            "sampy": float(sampy),
        }

    def manualCentringDone(self, manual_centring_procedure):
        try:
            motor_pos = manual_centring_procedure.get()
            if isinstance(motor_pos, gevent.GreenletExit):
                raise motor_pos
        except Exception:
            logging.exception("Could not complete manual centring")
            self.emitCentringFailed()
        else:
            self.emitProgressMessage("Moving sample to centred position...")
            self.emitCentringMoving()
            try:
                sample_centring.end()
            except Exception:
                logging.exception("Could not move to centred position")
                self.emitCentringFailed()

            # logging.info("EMITTING CENTRING SUCCESSFUL")
            self.centredTime = time.time()
            self.emitCentringSuccessful()
            self.emitProgressMessage("")

    def autoCentringDone(self, auto_centring_procedure):
        self.emitProgressMessage("")
        self.emit("newAutomaticCentringPoint", (-1, -1))

        try:
            res = auto_centring_procedure.get()
        except Exception:
            logging.error("Could not complete automatic centring")
            logging.getLogger("user_level_log").info("Automatic loop centring failed")
            self.emitCentringFailed()
            self.reject_centring()
        else:
            if res is None:
                logging.error("Could not complete automatic centring")
                logging.getLogger("user_level_log").info(
                    "Automatic loop centring failed"
                )
                self.emitCentringFailed()
                self.reject_centring()
            else:
                self.emitCentringSuccessful()
                self.accept_centring()
                logging.getLogger("user_level_log").info(
                    "Automatic loop centring successful"
                )

    def start_auto_centring(self, sample_info=None, loop_only=False):
        beam_pos_x, beam_pos_y = HWR.beamline.beam.get_beam_position_on_screen()

        self.set_phase("centring", wait=True)

        self.wait_ready(30)
        self.current_centring_procedure = sample_centring.start_auto(
            HWR.beamline.sample_view,
            {
                "phi": self.centringPhi,
                "phiy": self.centringPhiy,
                "sampx": self.centringSamplex,
                "sampy": self.centringSampley,
                "phiz": self.centringPhiz,
            },
            self.pixelsPerMmY,
            self.pixelsPerMmZ,
            beam_pos_x,
            beam_pos_y,
            chi_angle=float(self.chiAngle),
            msg_cb=self.emitProgressMessage,
            new_point_cb=lambda point: self.emit("newAutomaticCentringPoint", point),
        )

        self.current_centring_procedure.link(self.autoCentringDone)

        self.emitProgressMessage("Starting automatic centring procedure...")

    @task
    def moveToCentredPosition(self, centred_position):
        return self.move_motors(centred_position.as_dict())

    def image_clicked(self, x, y, xi, yi):
        logging.getLogger("user_level_log").info(
            "Centring click at, x: %s, y: %s" % (int(x), int(y))
        )
        sample_centring.user_click(x, y, True)

    def emitCentringStarted(self, method):
        self.currentCentringMethod = method
        self.emit("centringStarted", (method, False))
        logging.getLogger("user_level_log").info("Starting centring")

    def accept_centring(self):
        self.save_centring_positions()
        self.centringStatus["valid"] = True
        self.centringStatus["accepted"] = True
        self.emit("centringAccepted", (True, self.get_centring_status()))

        # save position in MD2 software
        self.save_centring_positions()

        self.log.info("DEBUG %s" % self.get_centring_status())
        logging.getLogger("user_level_log").info("Centring successful")

    def reject_centring(self):
        if self.current_centring_procedure:
            self.current_centring_procedure.kill()
        self.centringStatus = {"valid": False}
        self.emitProgressMessage("")
        self.emit("centringAccepted", (False, self.get_centring_status()))
        logging.getLogger("user_level_log").info("Centring cancelled")

    def emitCentringMoving(self):
        self.emit("centringMoving", ())

    def emitCentringFailed(self):
        self.centringStatus = {"valid": False}
        method = self.currentCentringMethod
        self.currentCentringMethod = None
        self.current_centring_procedure = None
        self.emit("centringFailed", (method, self.get_centring_status()))

    def emitCentringSuccessful(self):
        if self.current_centring_procedure is not None:
            curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
            self.centringStatus["endTime"] = curr_time
            self.centringStatus["motors"] = self.get_positions()
            centred_pos = self.current_centring_procedure.get()
            for role in self.centringStatus["motors"]:
                motor = self.get_object_by_role(role)
                try:
                    self.centringStatus["motors"][role] = centred_pos[motor]
                except KeyError:
                    continue

            self.centringStatus["method"] = self.currentCentringMethod
            self.centringStatus["valid"] = True

            method = self.currentCentringMethod
            self.emit("centringSuccessful", (method, self.get_centring_status()))
            self.currentCentringMethod = None
            self.current_centring_procedure = None
        else:
            self.log.debug(
                "MiniDiff: trying to emit centringSuccessful outside of a centring"
            )

    def emitProgressMessage(self, msg=None):
        # self.log.debug("%s: %s", self.name, msg)
        self.emit("progressMessage", (msg,))

    def get_centring_status(self):
        return copy.deepcopy(self.centringStatus)

    def get_positions(self):
        return {
            "phi": float(self.phiMotor.get_value()),
            "focus": float(self.focusMotor.get_value()),
            "phiy": float(self.phiyMotor.get_value()),
            "phiz": float(self.phizMotor.get_value()),
            "sampx": float(self.sampleXMotor.get_value()),
            "sampy": float(self.sampleYMotor.get_value()),
            "kappa": float(self.kappaMotor.get_value()) if self.kappaMotor else None,
            "kappa_phi": (
                float(self.kappaPhiMotor.get_value()) if self.kappaPhiMotor else None
            ),
            "zoom": float(self.zoomMotor.get_value()),
        }

    def move_motors(self, roles_positions_dict):
        motor = {
            "phi": self.phiMotor,
            "focus": self.focusMotor,
            "phiy": self.phiyMotor,
            "phiz": self.phizMotor,
            "sampx": self.sampleXMotor,
            "sampy": self.sampleYMotor,
            "kappa": self.kappaMotor,
            "kappa_phi": self.kappaPhiMotor,
            "zoom": self.zoomMotor,
        }

        for role, pos in roles_positions_dict.items():
            m = motor.get(role)
            if None not in (m, pos):
                m.set_value(pos)

        # TODO: remove this sleep, the motors states should
        # be MOVING since the beginning (or READY if move is
        # already finished)
        time.sleep(1)

        while not all(
            [m.get_state() == m.READY for m in motor.values() if m is not None]
        ):
            time.sleep(0.1)

    def take_snapshot(self, image_path_list: list) -> None:
        if len(image_path_list) > 0:
            if self.get_current_phase() != "Centring":
                use_custom_snapshot_routine = self.get_property(
                    "custom_snapshot_script_dir", False
                )

            if not use_custom_snapshot_routine:
                self.set_phase("Centring", wait=True, timeout=200)

        for image_path in image_path_list:
            snapshot_index = image_path_list.index(image_path)
            logging.getLogger("user_level_log").info(
                f"Taking {snapshot_index + 1} sample snapshot(s)"
            )
            HWR.beamline.sample_view.save_snapshot(path=image_path)
            # do not move 90 degrees if not needed
            if not self.in_plate_mode() and snapshot_index < len(image_path_list) - 1:
                self.phiMotor.set_value_relative(90, timeout=5)

    def snapshotsDone(self, snapshotsProcedure):
        HWR.beamline.sample_view.camera.forceUpdate = False

        try:
            self.centringStatus["images"] = snapshotsProcedure.get()
        except Exception:
            self.log.exception("MiniDiff: could not take crystal snapshots")
            self.emit("centringSnapshots", (False,))
            self.emitProgressMessage("")
        else:
            self.emit("centringSnapshots", (True,))
            self.emitProgressMessage("")
        self.emitProgressMessage("Sample is centred!")
        # self.emit('centringAccepted', (True,self.get_centring_status()))

    def simulateAutoCentring(self, sample_info=None):
        pass

    def wait_ready(self, timeout=None):
        pass

    def get_head_configuration(self) -> Union[GonioHeadConfiguration, None]:
        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        data = None

        if chip_def_fpath and os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "r") as _f:
                chip_def = json.load(_f)

                try:
                    data = GonioHeadConfiguration(**chip_def)
                except ValidationError:
                    self.log.exception("Validation error in %s" % chip_def_fpath)

        return data

    def set_head_configuration(self, str_data: str) -> None:
        data = json.loads(str_data)

        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        if chip_def_fpath and os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "w+") as _f:
                try:
                    GonioHeadConfiguration(**data)
                except ValidationError:
                    self.log.exception("Validation error in %s" % chip_def_fpath)

                _f.write(json.dumps(data, indent=4))

    def set_chip_layout(self, layout_name: str) -> True:
        data = self.get_head_configuration().dict()
        data["current"] = layout_name
        self.set_head_configuration(json.dumps(data))

    def run_script(self, script_cmd, wait=True):
        self._run_script(script_cmd)

        if wait:
            self._wait_ready(300)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""BeamstopMockup"""

from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["MXCuBE collaboration"]


class BeamstopMockup(HardwareObject):
    """
    Descrip. :
    """

    def __init__(self, name):
        """
        Descrip. :
        """
        HardwareObject.__init__(self, name)

        self.default_size = None
        self.default_distance = None
        self.default_direction = None

    def init(self):
        """
        Descrip. :
        """
        self.default_size = self.get_property("defaultBeamstopSize")
        self.default_distance = self.get_property("defaultBeamstopDistance")
        self.default_direction = self.get_property("defaultBeamstopDirection")

    def get_size(self):
        """
        Descrip. :
        """
        return self.default_size

    def set_distance(self, position):
        self.default_distance = position

    def get_distance(self):
        """
        Descrip. :
        """
        return self.default_distance

    def get_direction(self):
        """
        Descrip. :
        """
        return self.default_direction

    def re_emit_values(self):
        self.emit("beamstopDistanceChanged", (self.default_distance))



import xml.sax
from xml.sax.handler import ContentHandler

from mxcubecore.HardwareObjects.abstract.AbstractSampleChanger import (
    Container,
    Sample,
    SampleChanger,
    SampleChangerMode,
    SampleChangerState,
    gevent,
    sys,
)


class Pin(Sample):
    def __init__(self, basket, basket_no, sample_no):
        super(Pin, self).__init__(
            basket, Pin.get_sample_address(basket_no, sample_no), True
        )
        self._set_holder_length(22.0)

    def get_basket_no(self):
        return self.get_container().get_index() + 1

    def get_vial_no(self):
        return self.get_index() + 1

    @staticmethod
    def get_sample_address(basket_number, sample_number):
        return str(basket_number) + ":" + "%02d" % (sample_number)


class Basket(Container):
    __TYPE__ = "Puck"

    def __init__(self, container, number):
        super(Basket, self).__init__(
            self.__TYPE__, container, Basket.get_basket_address(number), True
        )
        for i in range(10):
            slot = Pin(self, number, i + 1)
            self._add_component(slot)

    @staticmethod
    def get_basket_address(basket_number):
        return str(basket_number)

    def clear_info(self):
        self.get_container()._reset_basket_info(self.get_index() + 1)
        self.get_container()._trigger_info_changed_event()


class XMLDataMatrixReadingHandler(ContentHandler):
    def __init__(self):
        ContentHandler.__init__(self)
        self.dataMatrixList = []
        self.basketDataMatrixList = [("", 4)] * 5
        self.dataMatrix = ""
        self.basketLocation = -1
        self.sampleLocation = -1
        self.flag = 0
        self.holder_length = 22.0

    def startElement(self, name, attrs):
        self.addBasketLocation = name == "BasketLocation"
        self.addSampleLocation = name == "Location"
        self.addDataMatrix = name == "DataMatrix"
        self.addFlag = name in ("SampleFlag", "BasketFlag")
        self.addHolderLength = name == "HolderLength"

    def characters(self, content):
        if self.addDataMatrix:
            self.dataMatrix = str(content)
        elif self.addBasketLocation:
            self.basketLocation = int(content)
        elif self.addSampleLocation:
            self.sampleLocation = int(content)
        elif self.addFlag:
            self.flag = int(content)
        elif self.addHolderLength:
            self.holder_length = float(content)

    def endElement(self, name):
        if name == "Sample":
            self.dataMatrixList.append(
                (
                    self.dataMatrix,
                    self.basketLocation,
                    self.sampleLocation,
                    self.flag,
                    self.holder_length,
                )
            )
            self.dataMatrix = ""
            self.basketLocation = -1
            self.sampleLocation = -1
            self.flag = 0
            self.holder_length = 22.0
        elif name == "Basket":
            self.basketDataMatrixList[self.sampleLocation - 1] = (
                self.dataMatrix,
                self.flag,
            )
            self.dataMatrix = ""
            self.basketLocation = -1
            self.sampleLocation = -1
            self.flag = 0
            self.holder_length = 22.0
        elif name == "BasketLocation":
            self.addBasketLocation = None
        elif name == "Location":
            self.addSampleLocation = None
        elif name == "DataMatrix":
            self.addDataMatrix = None
        elif name in ("SampleFlag", "BasketFlag"):
            self.addFlag = None
        elif name == "HolderLength":
            self.addHolderLength = None


class SC3(SampleChanger):
    __TYPE__ = "SC3"

    """
    Concrete implementation of SC3 Sample Changer
    """

    def __init__(self, *args, **kwargs):
        super(SC3, self).__init__(self.__TYPE__, True, *args, **kwargs)
        for i in range(5):
            basket = Basket(self, i + 1)
            self._add_component(basket)

    def init(self):
        for channel_name in ("_state", "_selected_basket", "_selected_sample"):
            setattr(self, channel_name, self.get_channel_object(channel_name))

        for command_name in (
            "_abort",
            "_getInfo",
            "_is_task_running",
            "_check_task_result",
            "_load",
            "_unload",
            "_chained_load",
            "_set_sample_charge",
            "_scan_basket",
            "_scan_samples",
            "_select_sample",
            "_select_basket",
            "_reset",
            "_reset_basket_info",
        ):
            setattr(self, command_name, self.get_command_object(command_name))

        SampleChanger.init(self)

    def get_sample_properties(self):
        return (Pin.__HOLDER_LENGTH_PROPERTY__,)

    def get_basket_list(self):
        basket_list = []
        for basket in self.get_components():
            if isinstance(basket, Basket):
                basket_list.append(basket)
        return basket_list

    # ########    TASKS    ########
    def _do_abort(self):
        self._abort()

    def _do_update_info(self):
        try:
            sxml = self._getInfo()
            handler = XMLDataMatrixReadingHandler()
            sxml = sxml.replace(' encoding="utf-16"', "")
            xml.sax.parseString(sxml, handler)
            sample_list = handler.dataMatrixList
            basket_list = handler.basketDataMatrixList
        except Exception:
            basket_list = [("", 4)] * 5
            sample_list = []
            for b in range(5):
                for s in range(10):
                    sample_list.append(("", b + 1, s + 1, 1, 22.0))

        for b in range(5):
            datamatrix = basket_list[b][0]
            if len(datamatrix) == 0:
                datamatrix = None
            flags = basket_list[b][1]

            present = (flags & 3) != 0
            scanned = (flags & 8) != 0
            basket = self.get_components()[b]
            basket._set_info(present, datamatrix, scanned)
        for s in sample_list:
            datamatrix = s[0]
            if len(datamatrix) == 0:
                datamatrix = None
            flags = s[3]
            present = (flags & 6) != 0
            scanned = present
            loaded = (flags & 8) != 0
            has_been_loaded = (flags & 16) != 0
            sample = self.get_component_by_address(Pin.get_sample_address(s[1], s[2]))
            sample._set_info(present, datamatrix, scanned)
            sample._set_loaded(loaded, has_been_loaded)
            sample._set_holder_length(s[4])
        self._update_selection()
        self._update_state()

    def _do_change_mode(self, mode):
        if mode == SampleChangerMode.Charging:
            self._execute_server_task(self._set_sample_charge, True)
        elif mode == SampleChangerMode.Normal:
            self._execute_server_task(self._set_sample_charge, False)

    def _do_scan(self, component, recursive):
        selected_basket = self.get_selected_component()
        if isinstance(component, Sample):
            if (selected_basket is None) or (
                selected_basket != component.get_container()
            ):
                self._do_select(component)
            self._execute_server_task(self._scan_samples, [component.get_index() + 1])

        elif isinstance(component, Container) and (
            component.get_type() == Basket.__TYPE__
        ):
            if recursive:
                self._execute_server_task(
                    self._scan_basket, (component.get_index() + 1)
                )
            else:
                if (selected_basket is None) or (selected_basket != component):
                    self._do_select(component)
                self._execute_server_task(self._scan_samples, (0,))
        elif isinstance(component, Container) and (
            component.get_type() == SC3.__TYPE__
        ):
            for basket in self.get_components():
                self._do_scan(basket, True)

    def _do_select(self, component):
        if isinstance(component, Sample):
            selected_basket = self.get_selected_component()
            if (selected_basket is None) or (
                selected_basket != component.get_container()
            ):
                self._execute_server_task(
                    self._select_basket, component.get_basket_no()
                )
            self._execute_server_task(self._select_sample, component.get_index() + 1)
        elif isinstance(component, Container) and (
            component.get_type() == Basket.__TYPE__
        ):
            self._execute_server_task(self._select_basket, component.get_index() + 1)

    def _do_load(self, sample=None):
        selected = self.get_selected_sample()
        if self.has_loaded_sample():
            if (sample is None) or (sample == self.get_loaded_sample()):
                raise Exception(
                    "The sample "
                    + str(self.get_loaded_sample().get_address())
                    + " is already loaded"
                )
            self._execute_server_task(
                self._chained_load,
                [
                    sample.get_basket_no(),
                    sample.get_vial_no(),
                    sample.get_holder_length(),
                ],
            )
        else:
            if sample is None:
                if selected is None:
                    raise Exception("No sample selected")
                else:
                    sample = selected
            elif (sample is not None) and (sample != selected):
                self._do_select(sample)
            self._execute_server_task(self._load, sample.get_holder_length())

    def _do_unload(self, sample_slot=None):
        if sample_slot is not None:
            self._do_select(sample_slot)
        self._execute_server_task(self._unload)

    def _do_reset(self):
        self._execute_server_task(self._reset)

    def clear_basket_info(self, basket):
        self._reset_basket_info(basket)

    # ########     PRIVATE   ########

    def _execute_server_task(self, method, *args):
        self._wait_device_ready(3.0)
        task_id = method(*args)
        ret = None
        if task_id is None:  # Reset
            while self._is_device_busy():
                gevent.sleep(0.1)
        else:
            while str(self._is_task_running(task_id)).lower() == "true":
                gevent.sleep(0.1)
            try:
                ret = self._check_task_result(task_id)
            except Exception:
                raise
        return ret

    def _update_state(self):
        try:
            state = self._read_state()
        except Exception:
            state = SampleChangerState.Unknown
        if state == SampleChangerState.Moving and self._is_device_busy(
            self.get_state()
        ):
            return
        self._set_state(state)

    def _read_state(self):
        state = str(self._state.get_value() or "").upper()
        state_converter = {
            "ALARM": SampleChangerState.Alarm,
            "FAULT": SampleChangerState.Fault,
            "MOVING": SampleChangerState.Moving,
            "STANDBY": SampleChangerState.Ready,
            "READY": SampleChangerState.Ready,
            "RUNNING": SampleChangerState.Moving,
            "LOADING": SampleChangerState.Charging,
            "INIT": SampleChangerState.Initializing,
        }
        return state_converter.get(state, SampleChangerState.Unknown)

    def _is_device_busy(self, state=None):
        if state is None:
            state = self._read_state()
        return state not in (
            SampleChangerState.Ready,
            SampleChangerState.Loaded,
            SampleChangerState.Alarm,
            SampleChangerState.Disabled,
            SampleChangerState.Fault,
            SampleChangerState.StandBy,
        )

    def _is_device_ready(self):
        state = self._read_state()
        return state in (SampleChangerState.Ready, SampleChangerState.Charging)

    def _wait_device_ready(self, timeout=None):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self._is_device_ready():
                gevent.sleep(0.01)

    def _update_selection(self):
        basket = None
        sample = None
        try:
            basket_no = self._selected_basket.get_value()
            if basket_no is not None and basket_no > 0 and basket_no <= 5:
                basket = self.get_component_by_address(
                    Basket.get_basket_address(basket_no)
                )
                sample_no = self._selected_sample.get_value()
                if sample_no is not None and sample_no > 0 and sample_no <= 10:
                    sample = self.get_component_by_address(
                        Pin.get_sample_address(basket_no, sample_no)
                    )
        except Exception:
            self.log.exception("")
        self._set_selected_component(basket)
        self._set_selected_sample(sample)


if __name__ == "__main__":

    def onStateChanged(state, former):
        print(("State Change:  " + str(former) + " => " + str(state)))

    def onInfoChanged():
        print("Info Changed")

    sc = SC3()
    sc.channel_type = "tango"
    sc.tangoname = "tango://pc445.embl.fr:17100/EMBL/MXSC/1#dbase=no"
    sc.channel_state = "State"
    sc.channel_selected_basket = "SelectedBasketLocation"
    sc.channel_selected_sample = "SelectedSampleLocation"
    sc.command_abort = "abort"
    sc.command_info = "GetInformationAsXML"
    sc.command_is_task_running = "isTaskRunning"
    sc.command_check_task_result = "checkTaskResult"
    sc.command_load = "LoadSample"
    sc.command_unload = "UnLoadSample"
    sc.command_chained_load = "ChainedLoadSample"
    sc.command_set_sample_charge = "SetBasketTransferMode"
    sc.command_scan_basket = "ScanBasketForDatamatrix"
    sc.command_scan_samples = "ScanSamplesForDatamatrix"
    sc.command_select_sample = "MoveSampleLocation"
    sc.command_select_basket = "MoveBasketLocation"
    sc.command_reset = "Reset"

    sc.connect(sc, sc.STATE_CHANGED_EVENT, onStateChanged)
    sc.connect(sc, sc.INFO_CHANGED_EVENT, onInfoChanged)
    sc.init()

    while True:
        try:
            sc.abort()
            sc.reset()
            sc.change_mode(SampleChangerMode.Normal)
            # or else sc.select(sc.get_component_by_address('1'), wait=True)
            sc.select("1", wait=True)
            sc.unload(wait=True)
            sc.load(None, wait=True)
            sc.load("3:04", wait=True)
            sc.unload("4:01", wait=True)
            sc.scan("1:01", wait=True)
            sc.scan("1:02", wait=True)
            sc.scan("1:03", wait=True)
            sc.scan("2:02", wait=True)
            sc.scan("3", wait=True)
            sc.scan("4", recursive=True, wait=True)
        except Exception:
            print((sys.exc_info()[1]))



import logging
import math
import os
import time

import gevent
import PyChooch
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.figure import Figure

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.TaskUtils import cleanup


class SOLEILEnergyScan(HardwareObject):
    def init(self):
        self.log.info(
            "#############################    EnergyScan: INIT HWOBJ  ###################"
        )
        self.ready_event = gevent.event.Event()
        self.scanning = None
        self.moving = None
        self.energyScanArgs = None
        self.archive_prefix = None
        self.energy2WavelengthConstant = None
        self.defaultWavelength = None
        self._element = None
        self._edge = None
        self.do_energy_scan = None
        try:
            self.defaultWavelengthChannel = self.get_channel_object(
                "default_wavelength"
            )
        except KeyError:
            self.defaultWavelengthChannel = None
        else:
            if self.defaultWavelengthChannel is not None:
                self.defaultWavelengthChannel.connect_signal(
                    "connected", self.sConnected
                )
                self.defaultWavelengthChannel.connect_signal(
                    "disconnected", self.sDisconnected
                )

        if self.defaultWavelengthChannel is None:
            # MAD beamline
            try:
                self.energyScanArgs = self.get_channel_object("escan_args")
            except KeyError:
                self.log.warning(
                    "EnergyScan: error initializing energy scan arguments (missing channel)"
                )
                self.energyScanArgs = None

            try:
                self.scanStatusMessage = self.get_channel_object("scanStatusMsg")
            except KeyError:
                self.scanStatusMessage = None
                self.log.warning(
                    "EnergyScan: energy messages will not appear (missing channel)"
                )
            else:
                self.connect(self.scanStatusMessage, "update", self.scanStatusChanged)

            try:
                self.do_energy_scan.connect_signal(
                    "commandReplyArrived", self.scanCommandFinished
                )
                self.do_energy_scan.connect_signal(
                    "commandBeginWaitReply", self.scanCommandStarted
                )
                self.do_energy_scan.connect_signal(
                    "commandFailed", self.scanCommandFailed
                )
                self.do_energy_scan.connect_signal(
                    "commandAborted", self.scanCommandAborted
                )
                self.do_energy_scan.connect_signal(
                    "commandReady", self.scanCommandReady
                )
                self.do_energy_scan.connect_signal(
                    "commandNotReady", self.scanCommandNotReady
                )
            except AttributeError as diag:
                self.log.warning(
                    "EnergyScan: error initializing energy scan (%s)" % str(diag)
                )
                self.do_energy_scan = Xanes  # .xanes(None, None) #None
            else:
                self.do_energy_scan.connect_signal("connected", self.sConnected)
                self.do_energy_scan.connect_signal("disconnected", self.sDisconnected)

            self.previousResolution = None
            self.lastResolution = None

            if HWR.beamline.lims is None:
                self.log.warning(
                    "EnergyScan: you should specify the database hardware object"
                )
            self.scanInfo = None

            self.cryostreamHO = self.get_object_by_role("cryostream")
            if self.cryostreamHO is None:
                self.log.warning(
                    "EnergyScan: you should specify the cryo stream hardware object"
                )

            self.machcurrentHO = self.get_object_by_role("machcurrent")
            if self.machcurrentHO is None:
                self.log.warning(
                    "EnergyScan: you should specify the machine current hardware object"
                )

            self.fluodetectorHO = self.get_object_by_role("fluodetector")
            if self.fluodetectorHO is None:
                self.log.warning(
                    "EnergyScan: you should specify the fluorescence detector hardware object"
                )

            try:
                # self.moveEnergy.connect_signal('commandReplyArrived', self.moveEnergyCmdFinished)
                # self.moveEnergy.connect_signal('commandBeginWaitReply', self.moveEnergyCmdStarted)
                # self.moveEnergy.connect_signal('commandFailed', self.moveEnergyCmdFailed)
                # self.moveEnergy.connect_signal('commandAborted', self.moveEnergyCmdAborted)
                self.moveEnergy.connect_signal("commandReady", self.moveEnergyCmdReady)
                self.moveEnergy.connect_signal(
                    "commandNotReady", self.moveEnergyCmdNotReady
                )
            except AttributeError as diag:
                self.log.warning(
                    "EnergyScan: error initializing move energy (%s)" % str(diag)
                )
                self.moveEnergy = None

            if HWR.beamline.energy is not None:
                HWR.beamline.energy.connect("valueChanged", self.energyPositionChanged)
                HWR.beamline.energy.connect("stateChanged", self.energyStateChanged)
                HWR.beamline.energy.connect("limitsChanged", self.energyLimitsChanged)
            if HWR.beamline.resolution is None:
                self.log.warning(
                    "EnergyScan: no resolution motor (unable to restore it after moving the energy)"
                )
            else:
                HWR.beamline.resolution.connect(
                    "valueChanged", self.resolutionPositionChanged
                )

        self.thEdgeThreshold = self.get_property("theoritical_edge_threshold")
        if self.thEdgeThreshold is None:
            self.thEdgeThreshold = 0.01

        if self.is_connected():
            self.sConnected()
        self.log.info(
            "#############################    EnergyScan: INIT HWOBJ IS FINISHED ###################"
        )

    def is_connected(self):
        if self.defaultWavelengthChannel is not None:
            # single wavelength beamline
            try:
                return self.defaultWavelengthChannel.is_connected()
            except Exception:
                return False
        else:
            try:
                return self.do_energy_scan.is_connected()
            except Exception:
                return False

    def resolutionPositionChanged(self, res):
        self.lastResolution = res

    def energyStateChanged(self, state):
        if state == HWR.beamline.energy.READY:
            if HWR.beamline.resolution is not None:
                HWR.beamline.resolution.dist2res()

    # Handler for spec connection
    def sConnected(self):
        self.emit("connected", ())

    # Handler for spec disconnection
    def sDisconnected(self):
        self.emit("disconnected", ())

    def setElement(self):
        self.log.debug("EnergyScan: setElement")
        self.emit("setElement", (self._element, self._edge))

    def newPoint(self, x, y):
        self.log.debug("EnergyScan:newPoint")
        logging.info("EnergyScan newPoint %s, %s" % (x, y))
        self.emit("addNewPoint", (x, y))
        self.emit("newScanPoint", (x, y))

    def newScan(self, scanParameters):
        self.log.debug("EnergyScan:newScan")
        self.emit("newScan", (scanParameters,))

    # # Energy scan commands
    # def canScanEnergy(self):
    #     if not self.is_connected():
    #         return False
    #     if self.energy2WavelengthConstant is None or self.energyScanArgs is None:
    #         return False
    #     return self.do_energy_scan is not None

    def start_energy_scan(
        self, element, edge, directory, prefix, session_id=None, blsample_id=None
    ):
        self._element = element
        self._edge = edge
        self.log.debug(
            "EnergyScan: starting energy scan %s, %s" % (self._element, self._edge)
        )
        self.setElement()
        self.xanes = Xanes.xanes(
            self,
            element,
            edge,
            directory,
            prefix,
            session_id,
            blsample_id,
            plot=False,
            test=False,
        )
        self.scanInfo = {
            "sessionId": session_id,
            "blSampleId": blsample_id,
            "element": element,
            "edgeEnergy": edge,
        }
        if self.fluodetectorHO is not None:
            self.scanInfo["fluorescenceDetector"] = self.fluodetectorHO.username
        if not os.path.isdir(directory):
            self.log.debug("EnergyScan: creating directory %s" % directory)
            try:
                os.makedirs(directory)
            except OSError as diag:
                self.log.error(
                    "EnergyScan: error creating directory %s (%s)"
                    % (directory, str(diag))
                )
                self.emit("scanStatusChanged", ("Error creating directory",))
                return False

        scanParameter = {}
        scanParameter["title"] = "Energy Scan"
        scanParameter["xlabel"] = "Energy in keV"
        scanParameter["ylabel"] = "Normalized counts"
        self.newScan(scanParameter)

        # try:
        # curr=self.energyScanArgs.get_value()
        # except:
        # self.log.exception('EnergyScan: error getting energy scan parameters')
        # self.emit('scanStatusChanged', ("Error getting energy scan parameters",))
        # return False
        # try:
        # curr["escan_dir"]=directory
        # curr["escan_prefix"]=prefix
        # except TypeError:
        curr = {}
        curr["escan_dir"] = directory
        curr["escan_prefix"] = prefix

        self.archive_prefix = prefix

        try:
            # self.energyScanArgs.set_value(curr)
            self.log.debug(
                "EnergyScan: current energy scan parameters (%s, %s, %s, %s)"
                % (element, edge, directory, prefix)
            )
        except Exception:
            self.log.exception("EnergyScan: error setting energy scan parameters")
            self.emit("scanStatusChanged", ("Error setting energy scan parameters",))
            return False
        try:
            # self.do_energy_scan("%s %s" % (element,edge))
            self.scanCommandStarted()
            self.xanes.scan()  # start() #scan()
            self.scanCommandFinished("success")
        except Exception:
            self.log.exception("EnergyScan: problem calling sequence")
            self.emit("scanStatusChanged", ("Error problem spec macro",))
            return False
        return True

    def cancelEnergyScan(self, *args):
        logging.info("SOLEILEnergyScan: canceling the scan")
        if self.scanning:
            # self.do_energy_scan.abort()
            self.xanes.abort()
            self.ready_event.set()

    def scanCommandReady(self):
        if not self.scanning:
            self.emit("energyScanReady", (True,))

    def scanCommandNotReady(self):
        if not self.scanning:
            self.emit("energyScanReady", (False,))

    def scanCommandStarted(self, *args):
        self.scanInfo["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = True
        self.emit("energyScanStarted", ())

    def scanCommandFailed(self, *args):
        self.scanInfo["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.scanning = False
        self.storeEnergyScan()
        self.emit("energyScanFailed", ())
        self.ready_event.set()

    def scanCommandAborted(self, *args):
        self.emit("energyScanFailed", ())
        self.ready_event.set()

    def scanCommandFinished(self, result, *args):
        self.log.debug("EnergyScan: energy scan result is %s" % result)
        with cleanup(self.ready_event.set):
            self.scanInfo["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
            self.log.debug("EnergyScan: energy scan result is %s" % result)
            self.scanning = False
            if result == -1:
                self.storeEnergyScan()
                self.emit("energyScanFailed", ())
                return

            try:
                t = float(result["transmissionFactor"])
            except Exception:
                pass
            else:
                self.scanInfo["transmissionFactor"] = t
            try:
                et = float(result["exposureTime"])
            except Exception:
                pass
            else:
                self.scanInfo["exposureTime"] = et
            try:
                se = float(result["startEnergy"])
            except Exception:
                pass
            else:
                self.scanInfo["startEnergy"] = se
            try:
                ee = float(result["endEnergy"])
            except Exception:
                pass
            else:
                self.scanInfo["endEnergy"] = ee

            try:
                bsX = float(result["beamSizeHorizontal"])
            except Exception:
                pass
            else:
                self.scanInfo["beamSizeHorizontal"] = bsX

            try:
                bsY = float(result["beamSizeVertical"])
            except Exception:
                pass
            else:
                self.scanInfo["beamSizeVertical"] = bsY

            try:
                self.thEdge = float(result["theoreticalEdge"]) / 1000.0
            except Exception:
                pass

            self.emit("energyScanFinished", (self.scanInfo,))
            time.sleep(0.1)
            self.emit("energyScanFinished2", (self.scanInfo,))

    def do_chooch(self, elt, edge, scanArchiveFilePrefix, scanFilePrefix):
        # symbol = "_".join((elt, edge))
        # scanArchiveFilePrefix = "_".join((scanArchiveFilePrefix, symbol))

        # i = 1
        # while os.path.isfile(os.path.extsep.join((scanArchiveFilePrefix + str(i), "raw"))):
        # i = i + 1

        # scanArchiveFilePrefix = scanArchiveFilePrefix + str(i)
        # archiveRawScanFile=os.path.extsep.join((scanArchiveFilePrefix, "raw"))
        rawScanFile = os.path.extsep.join((scanFilePrefix, "raw"))
        scanFile = os.path.extsep.join((scanFilePrefix, "efs"))
        logging.info(
            "SOLEILEnergyScan do_chooch rawScanFile %s, scanFile %s"
            % (rawScanFile, scanFile)
        )
        # if not os.path.exists(os.path.dirname(scanArchiveFilePrefix)):
        # os.makedirs(os.path.dirname(scanArchiveFilePrefix))

        # try:
        # f=open(rawScanFile, "w")
        # pyarch_f=open(archiveRawScanFile, "w")
        # except:
        # self.log.exception("could not create raw scan files")
        # self.storeEnergyScan()
        # self.emit("energyScanFailed", ())
        # return
        # else:
        # scanData = []

        # if scanObject is None:
        # raw_data_file = os.path.join(os.path.dirname(scanFilePrefix), 'data.raw')
        # try:
        # raw_file = open(raw_data_file, 'r')
        # except:
        # self.storeEnergyScan()
        # self.emit("energyScanFailed", ())
        # return

        # for line in raw_file.readlines()[2:]:
        # (x, y) = line.split('\t')
        # x = float(x.strip())
        # y = float(y.strip())
        # x = x < 1000 and x*1000.0 or x
        # scanData.append((x, y))
        # f.write("%f,%f\r\n" % (x, y))
        # pyarch_f.write("%f,%f\r\n"% (x, y))
        # else:
        # for i in range(len(scanObject.x)):
        # x = float(scanObject.x[i])
        # x = x < 1000 and x*1000.0 or x
        # y = float(scanObject.y[i])
        # scanData.append((x, y))
        # f.write("%f,%f\r\n" % (x, y))
        # pyarch_f.write("%f,%f\r\n"% (x, y))

        # f.close()
        # pyarch_f.close()
        # self.scanInfo["scanFileFullPath"]=str(archiveRawScanFile)
        scanData = self.xanes.raw
        logging.info("scanData %s" % scanData)
        logging.info("PyChooch file %s" % PyChooch.__file__)
        pk, fppPeak, fpPeak, ip, fppInfl, fpInfl, chooch_graph_data = PyChooch.calc(
            scanData, elt, edge, scanFile
        )
        rm = (pk + 30) / 1000.0
        pk = pk / 1000.0
        savpk = pk
        ip = ip / 1000.0
        comm = ""

        self.thEdge = self.xanes.e_edge
        self.log.info(
            "th. Edge %s ; chooch results are pk=%f, ip=%f, rm=%f"
            % (self.thEdge, pk, ip, rm)
        )
        logging.info("math.fabs(self.thEdge - ip) %s" % math.fabs(self.thEdge - ip))
        logging.info("self.thEdgeThreshold %s" % self.thEdgeThreshold)

        self.log.warning(
            "EnergyScan: calculated peak (%f) is more that 20eV %s the theoretical value (%f). Please check your scan and choose the energies manually"
            % (
                savpk,
                (self.thEdge - ip) < self.thEdgeThreshold and "below" or "above",
                self.thEdge,
            )
        )

        if math.fabs(self.thEdge - ip) > self.thEdgeThreshold:
            logging.info("Theoretical edge too different from the one just determined")
            pk = 0
            ip = 0
            rm = self.thEdge + 0.03
            comm = (
                "Calculated peak (%f) is more that 10eV away from the theoretical value (%f). Please check your scan"
                % (savpk, self.thEdge)
            )

            self.log.warning(
                "EnergyScan: calculated peak (%f) is more that 20eV %s the theoretical value (%f). Please check your scan and choose the energies manually"
                % (
                    savpk,
                    (self.thEdge - ip) < self.thEdgeThreshold and "below" or "above",
                    self.thEdge,
                )
            )

        archiveEfsFile = os.path.extsep.join((scanArchiveFilePrefix, "efs"))

        logging.info("archiveEfsFile %s" % archiveEfsFile)

        # Check access to archive directory
        dirname = os.path.dirname(archiveEfsFile)
        if not os.path.exists(dirname):
            try:
                os.makedirs(dirname)
                logging.getLogger("user_level_log").info(
                    "Chooch. Archive path (%s) created" % dirname
                )
            except OSError:
                logging.getLogger("user_level_log").error(
                    "Chooch. Archive path is not accessible (%s)" % dirname
                )
                return None
            except Exception:
                self.user_log.error("Error creating archive path")
                self.log.exception("Error creating archive path")

                return None
        else:
            if not os.path.isdir(dirname):
                logging.getLogger("user_level_log").error(
                    "Chooch. Archive path does not seem to be a valid directory (%s)"
                    % dirname
                )
                return None

        try:
            fi = open(scanFile)
            fo = open(archiveEfsFile, "w")
        except Exception:
            self.user_log.exception("")
            self.storeEnergyScan()
            self.emit("energyScanFailed", ())
            return None
        else:
            fo.write(fi.read())
            fi.close()
            fo.close()

        logging.info("archive saved")
        self.scanInfo["peakEnergy"] = pk
        self.scanInfo["inflectionEnergy"] = ip
        self.scanInfo["remoteEnergy"] = rm
        self.scanInfo["peakFPrime"] = fpPeak
        self.scanInfo["peakFDoublePrime"] = fppPeak
        self.scanInfo["inflectionFPrime"] = fpInfl
        self.scanInfo["inflectionFDoublePrime"] = fppInfl
        self.scanInfo["comments"] = comm
        logging.info("self.scanInfo %s" % self.scanInfo)

        logging.info("chooch_graph_data %s" % str(chooch_graph_data))
        chooch_graph_x, chooch_graph_y1, chooch_graph_y2 = zip(*chooch_graph_data)
        chooch_graph_x = list(chooch_graph_x)
        logging.info("chooch_graph_x %s" % str(chooch_graph_x))
        for i in range(len(chooch_graph_x)):
            chooch_graph_x[i] = chooch_graph_x[i] / 1000.0

        self.log.info("<chooch> Saving png")
        # prepare to save png files
        title = "%10s  %6s  %6s\n%10s  %6.2f  %6.2f\n%10s  %6.2f  %6.2f" % (
            "energy",
            "f'",
            "f''",
            pk,
            fpPeak,
            fppPeak,
            ip,
            fpInfl,
            fppInfl,
        )
        fig = Figure(figsize=(15, 11))
        ax = fig.add_subplot(211)
        ax.set_title("%s\n%s" % (scanFile, title))
        ax.grid(True)
        ax.plot(*(zip(*scanData)), **{"color": "black"})
        ax.set_xlabel("Energy")
        ax.set_ylabel("MCA counts")
        ax2 = fig.add_subplot(212)
        ax2.grid(True)
        ax2.set_xlabel("Energy")
        ax2.set_ylabel("")
        handles = []
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y1, color="blue"))
        handles.append(ax2.plot(chooch_graph_x, chooch_graph_y2, color="red"))
        canvas = FigureCanvasAgg(fig)

        escan_png = os.path.extsep.join((scanFilePrefix, "png"))
        escan_archivepng = os.path.extsep.join((scanArchiveFilePrefix, "png"))
        self.scanInfo["jpegChoochFileFullPath"] = str(escan_archivepng)
        try:
            self.log.info(
                "Rendering energy scan and Chooch graphs to PNG file : %s", escan_png
            )
            canvas.print_figure(escan_png, dpi=80)
        except Exception:
            self.log.exception("could not print figure")
        try:
            self.log.info(
                "Saving energy scan to archive directory for ISPyB : %s",
                escan_archivepng,
            )
            canvas.print_figure(escan_archivepng, dpi=80)
        except Exception:
            self.log.exception("could not save figure")

        self.storeEnergyScan()
        self.scanInfo = None

        self.log.info("<chooch> returning")
        self.emit(
            "chooch_finished",
            (
                pk,
                fppPeak,
                fpPeak,
                ip,
                fppInfl,
                fpInfl,
                rm,
                chooch_graph_x,
                chooch_graph_y1,
                chooch_graph_y2,
                title,
            ),
        )
        self.choochResults = (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        )
        return (
            pk,
            fppPeak,
            fpPeak,
            ip,
            fppInfl,
            fpInfl,
            rm,
            chooch_graph_x,
            chooch_graph_y1,
            chooch_graph_y2,
            title,
        )

    def scanStatusChanged(self, status):
        self.emit("scanStatusChanged", (status,))

    def storeEnergyScan(self):
        # self.xanes.saveDat()
        self.xanes.saveRaw()
        self.xanes.saveResults()

        # if HWR.beamline.lims is None:
        # return
        # try:
        # session_id=int(self.scanInfo['sessionId'])
        # except:
        # return
        # gevent.spawn(StoreEnergyScanThread, HWR.beamline.lims,self.scanInfo)
        logging.info("SOLEILEnergyScan storeEnergyScan OK")
        # self.storeScanThread.start()

    def updateEnergyScan(self, scan_id, jpeg_scan_filename):
        pass

    # Move energy commands

    def get_current_energy(self):
        if HWR.beamline.energy is not None:
            try:
                return HWR.beamline.energy.get_value()
            except Exception:
                self.log.exception("EnergyScan: couldn't read energy")
                return None
        elif (
            self.energy2WavelengthConstant is not None
            and self.defaultWavelength is not None
        ):
            return self.energy2wavelength(self.defaultWavelength)

        return None

    def get_value(self):
        return self.get_current_energy()

    def getEnergyLimits(self):
        lims = None
        if HWR.beamline.energy is not None:
            if HWR.beamline.energy.is_ready():
                lims = HWR.beamline.energy.get_limits()
        return lims

    def get_wavelength(self):
        if HWR.beamline.energy is not None:
            try:
                return self.energy2wavelength(HWR.beamline.energy.get_value())
            except Exception:
                self.log.exception("EnergyScan: couldn't read energy")
                return None
        else:
            return self.defaultWavelength

    def getWavelengthLimits(self):
        limits = None
        if HWR.beamline.energy is not None:
            if HWR.beamline.energy.is_ready():
                limits = HWR.beamline.energy.get_wavelength_limits()
        return limits

    def startMoveEnergy(self, value, wait=True):
        self.log.info("Moving energy to (%s)" % value)
        try:
            value = float(value)
        except (TypeError, ValueError) as diag:
            self.log.error("EnergyScan: invalid energy (%s)" % value)
            return False

        try:
            curr_energy = HWR.beamline.energy.get_value()
        except Exception:
            self.log.exception("EnergyScan: couldn't get current energy")
            curr_energy = None

        if value != curr_energy:
            self.log.info("Moving energy: checking limits")
            try:
                lims = HWR.beamline.energy.get_limits()
            except Exception:
                self.log.exception("EnergyScan: couldn't get energy limits")
                in_limits = False
            else:
                in_limits = value >= lims[0] and value <= lims[1]

            if in_limits:
                self.log.info("Moving energy: limits ok")
                self.previousResolution = None
                if HWR.beamline.resolution is not None:
                    try:
                        self.previousResolution = HWR.beamline.resolution.get_value()
                    except Exception:
                        self.log.exception(
                            "EnergyScan: couldn't get current resolution"
                        )
                self.moveEnergyCmdStarted()

                def change_egy():
                    try:
                        self.moveEnergy(value, wait=True)
                    except Exception:
                        self.moveEnergyCmdFailed()
                    else:
                        self.moveEnergyCmdFinished(True)

                if wait:
                    change_egy()
                else:
                    gevent.spawn(change_egy)
            else:
                self.log.error(
                    "EnergyScan: energy (%f) out of limits (%s)" % (value, lims)
                )
                return False
        else:
            return None

        return True

    def set_wavelength(self, value, wait=True):
        energy_val = self.energy2wavelength(value)
        if energy_val is None:
            self.log.error("EnergyScan: unable to convert wavelength to energy")
            return False
        return self.startMoveEnergy(energy_val, wait)

    def cancelMoveEnergy(self):
        self.moveEnergy.abort()

    def energy2wavelength(self, val):
        if self.energy2WavelengthConstant is None:
            return None
        try:
            other_val = self.energy2WavelengthConstant / val
        except ZeroDivisionError:
            other_val = None
        return other_val

    def energyPositionChanged(self, pos):
        wav = self.energy2wavelength(pos)
        if wav is not None:
            self.emit("energyChanged", (pos, wav))
            self.emit("valueChanged", (pos,))

    def energyLimitsChanged(self, limits):
        self.emit("energyLimitsChanged", (limits,))
        wav_limits = (
            self.energy2wavelength(limits[1]),
            self.energy2wavelength(limits[0]),
        )
        if wav_limits[0] is not None and wav_limits[1] is not None:
            self.emit("wavelengthLimitsChanged", (wav_limits,))
        else:
            self.emit("wavelengthLimitsChanged", (None,))

    def moveEnergyCmdReady(self):
        if not self.moving:
            self.emit("moveEnergyReady", (True,))

    def moveEnergyCmdNotReady(self):
        if not self.moving:
            self.emit("moveEnergyReady", (False,))

    def moveEnergyCmdStarted(self):
        self.moving = True
        self.emit("moveEnergyStarted", ())

    def moveEnergyCmdFailed(self):
        self.moving = False
        self.emit("moveEnergyFailed", ())

    def moveEnergyCmdAborted(self):
        pass
        # self.moving = False
        # self.emit('moveEnergyFailed', ())

    def moveEnergyCmdFinished(self, result):
        self.moving = False
        self.emit("moveEnergyFinished", ())

    def getPreviousResolution(self):
        return (self.previousResolution, self.lastResolution)

    def restoreResolution(self):
        if HWR.beamline.resolution is not None:
            if self.previousResolution is not None:
                try:
                    HWR.beamline.resolution.set_value(self.previousResolution)
                except Exception:
                    return (False, "Error trying to move the detector")
                else:
                    return (True, None)
            else:
                return (False, "Unknown previous resolution")
        else:
            return (False, "Resolution motor not defined")

    # Elements commands
    def get_elements(self):
        elements = []
        try:
            for el in self["elements"]:
                elements.append({"symbol": el.symbol, "energy": el.energy})
        except IndexError:
            pass
        return elements

    # # Mad energies commands
    # def getDefaultMadEnergies(self):
    #     energies = []
    #     try:
    #         for el in self["mad"]:
    #             energies.append([float(el.energy), el.directory])
    #     except IndexError:
    #         pass
    #     return energies


def StoreEnergyScanThread(db_conn, scan_info):
    return
    scanInfo = dict(scan_info)
    dbConnection = db_conn

    blsampleid = scanInfo["blSampleId"]
    scanInfo.pop("blSampleId")
    db_status = dbConnection.storeEnergyScan(scanInfo)
    if blsampleid is not None:
        try:
            energyscanid = int(db_status["energyScanId"])
        except Exception:
            pass
        else:
            asoc = {"blSampleId": blsampleid, "energyScanId": energyscanid}
            dbConnection.associateBLSampleAndEnergyScan(asoc)



import logging

import gevent.event

from mxcubecore.HardwareObjects import edna_test_data
from mxcubecore.HardwareObjects.abstract.AbstractCharacterisation import (
    AbstractCharacterisation,
)
from mxcubecore.HardwareObjects.XSDataMXCuBEv1_3 import XSDataResultMXCuBE


class SOLEILEDNACharacterisationMockup(AbstractCharacterisation):
    def __init__(self, name):
        super(SOLEILEDNACharacterisationMockup, self).__init__(name)
        self.processing_done_event = gevent.event.Event()

    def get_html_report(self, edna_result):
        html_report = "/home/blissadm/mxcube/extras/result.html"
        return html_report

    def execute_command(self, command_name, *args, **kwargs):
        wait = kwargs.get("wait", True)
        cmd_obj = self.get_command_object(command_name)
        return cmd_obj(*args, wait=wait)

    def get_beam_size(self):
        return (10, 5)

    def input_from_params(self, data_collection, char_params):
        return char_params

    def characterise(self, edna_input):
        msg = "Starting MOCKUP Analisys"
        logging.getLogger("queue_exec").info(msg)

        self.processing_done_event.set()
        self.result = XSDataResultMXCuBE.parseString(edna_test_data.EDNA_RESULT_DATA)

        return self.result

    def is_running(self):
        return not self.processing_done_event.is_set()



#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import subprocess
import time

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.XSDataAutoprocv1_0 import XSDataAutoprocInput
from mxcubecore.HardwareObjects.XSDataCommon import (
    XSDataDouble,
    XSDataFile,
    XSDataInteger,
    XSDataString,
)

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class OfflineProcessingMockup(HardwareObject):
    """Hwobj assembles input xml and launches EDNAproc autoprocessing"""

    def __init__(self, name):
        HardwareObject.__init__(self, name)
        self.result = None
        self.autoproc_programs = []

    def init(self):
        try:
            self.autoproc_programs = self["programs"]
        except KeyError:
            self.print_log("AutoProcessing: no autoprocessing program defined.")

    def execute_autoprocessing(
        self, process_event, params_dict, frame_number, run_processing=True
    ):
        """Method called from collection hwobj after successful collection.

        :param process_event: processing type (after, before, image)
        :type process_event: str
        :param params_dict: collection parameters
        :type params_dict: dict
        :param frame_number: frame number
        :type frame_number: int
        :param run_processing: True = run processing or
                               False = create just input file
        :type run_processing: bool
        """
        self.autoproc_procedure(
            process_event, params_dict, frame_number, run_processing
        )

    def autoproc_procedure(
        self, process_event, params_dict, frame_number, run_processing=True
    ):
        """
        Main autoprocessing procedure. At the beginning correct event (defined
        in xml) is found. If the event is executable then accordingly to the
        event type (image, after) then the sequence is executed:
        Implemented tasks:
           - after : Main autoprocessing procedure
                     1. Input file is generated with create_autoproc_input
                        Input file has a name template
                        "edna-autoproc-input-%Y%m%d_%H%M%S.xml".
                     2. Then it waits for XDS.INP directory and if it exists
                        then creates input file
                     3. edna_autoprocessing.sh script is
                        executed with parameters:
                        - arg1 : generated xml file
                        - arg2 : process dir
                     4. script executes EDNA EDPluginControlEDNAproc
           - image : Thumbnail generation for first and last image
                     1. No input file is generated
                     2. edna_thumbnails.sh script is executed with parameters:
                        - arg1 : image base dir (place where thumb
                                 will be generated)
                        - arg2 : file name

        :param process_event: processing type (after, before, image)
        :type process_event: str
        :param params_dict: collection parameters
        :type params_dict: dict
        :param frame_number: frame number
        :type frame_number: int
        :param run_processing: True = run processing or
                               False = create just input file
        :type run_processing: bool
        """
        for program in self.autoproc_programs:
            if process_event == program.get_property("event"):
                executable = program.get_property("executable")

                will_execute = True
                if process_event == "after":
                    will_execute = run_processing
                    end_of_line_to_execute = " %s %s " % (
                        params_dict["xds_dir"],
                        params_dict.get("collection_id"),
                    )
                elif process_event == "image":
                    filename = params_dict["fileinfo"]["template"] % frame_number
                    end_of_line_to_execute = " %s %s/%s" % (
                        params_dict["fileinfo"]["archive_directory"],
                        params_dict["fileinfo"]["directory"],
                        filename,
                    )
                if will_execute:
                    subprocess.Popen(
                        str(executable + end_of_line_to_execute),
                        shell=True,
                        stdin=None,
                        stdout=None,
                        stderr=None,
                        close_fds=True,
                    )

    def create_autoproc_input(self, event, params):
        """Creates processing input xml

        :param event: processing type (after, before, image)
        :type event: str
        :param params: collection parameters
        :type params: dict
        """
        xds_input_file_wait_timeout = 20
        xds_input_file_wait_resolution = 1

        file_name_timestamp = time.strftime("%Y%m%d_%H%M%S")

        autoproc_path = params.get("xds_dir")
        autoproc_xds_filename = os.path.join(autoproc_path, "XDS.INP")
        autoproc_input_filename = os.path.join(
            autoproc_path, "edna-autoproc-input-%s.xml" % file_name_timestamp
        )
        autoproc_output_file_name = os.path.join(
            autoproc_path, "edna-autoproc-results-%s.xml" % file_name_timestamp
        )

        autoproc_input = XSDataAutoprocInput()
        autoproc_xds_file = XSDataFile()
        autoproc_xds_file.setPath(XSDataString(autoproc_xds_filename))
        autoproc_input.setInput_file(autoproc_xds_file)

        autoproc_output_file = XSDataFile()
        autoproc_output_file.setPath(XSDataString(autoproc_output_file_name))
        autoproc_input.setOutput_file(autoproc_output_file)

        autoproc_input.setData_collection_id(XSDataInteger(params.get("collection_id")))
        residues_num = float(params.get("residues", 0))
        if residues_num != 0:
            autoproc_input.setNres(XSDataDouble(residues_num))
        space_group = params.get("sample_reference").get("spacegroup", "")
        if len(space_group) > 0:
            autoproc_input.setSpacegroup(XSDataString(space_group))
        unit_cell = params.get("sample_reference").get("cell", "")
        if len(unit_cell) > 0:
            autoproc_input.setUnit_cell(XSDataString(unit_cell))

        autoproc_input.setCc_half_cutoff(XSDataDouble(18.0))

        # Maybe we have to check if directory is there.
        # Maybe create dir with mxcube
        xds_appeared = False
        wait_xds_start = time.time()
        logging.debug(
            "AutoprocessingMockup: Waiting for XDS.INP "
            + "file: %s" % autoproc_xds_filename
        )
        while (
            not xds_appeared
            and time.time() - wait_xds_start < xds_input_file_wait_timeout
        ):
            if (
                os.path.exists(autoproc_xds_filename)
                and os.stat(autoproc_xds_filename).st_size > 0
            ):
                xds_appeared = True
                logging.debug(
                    "AutoprocessingMockup: XDS.INP file is there, size={0}".format(
                        os.stat(autoproc_xds_filename).st_size
                    )
                )
            else:
                os.system("ls %s> /dev/null" % (os.path.dirname(autoproc_path)))
                gevent.sleep(xds_input_file_wait_resolution)
        if not xds_appeared:
            logging.error(
                "AutoprocessingMockup: XDS.INP file %s failed " % autoproc_xds_filename
                + "to appear after %d seconds" % xds_input_file_wait_timeout
            )
            return None, False

        autoproc_input.exportToFile(autoproc_input_filename)

        return autoproc_input_filename, True



import logging

import numpy
from PyTango import DeviceProxy

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.Command.Tango import TangoCommand
from mxcubecore.HardwareObjects.abstract.AbstractMotor import MotorStates


class TangoDCMotor(HardwareObject):
    MOVESTARTED = 0
    NOTINITIALIZED = 0
    UNUSABLE = 0
    READY = 2
    MOVING = 4
    ONLIMIT = 1

    stateDict = {
        "UNKNOWN": 0,
        "OFF": 0,
        "ALARM": 1,
        "FAULT": 1,
        "STANDBY": 2,
        "RUNNING": 4,
        "MOVING": 4,
        "ON": 2,
    }

    def __init__(self, name):
        # State values as expected by Motor bricks

        super().__init__(name)
        self.GUIstep = 0.1
        self.motor_states = MotorStates()

    def _init(self):
        self.positionValue = 0.0
        self.stateValue = "UNKNOWN"

        threshold = self.get_property("threshold")
        self.threshold = (
            0.0018  # default value. change it with property threshold in xml
        )

        self.old_value = 0.0
        self.tangoname = self.get_property("tangoname")
        self.motor_name = self.get_property("motor_name")
        self.ho = DeviceProxy(self.tangoname)

        try:
            self.dataType = self.get_property("datatype")
            if self.dataType is None:
                self.dataType = "float"
        except Exception:
            self.dataType = "float"

        if threshold is not None:
            try:
                self.threshold = float(threshold)
            except Exception:
                self.log.exception("")

        self.set_is_ready(True)
        try:
            self.limitsCommand = self.get_command_object("limits")
        except KeyError:
            self.limitsCommand = None
        self.positionChan = self.get_channel_object(
            "position"
        )  # utile seulement si positionchan n'est pas defini dans le code
        self.stateChan = self.get_channel_object(
            "state"
        )  # utile seulement si statechan n'est pas defini dans le code

        self.positionChan.connect_signal("update", self.positionChanged)
        self.stateChan.connect_signal("update", self.motorStateChanged)

    def positionChanged(self, value):
        self.positionValue = value
        if abs(float(value) - self.old_value) > self.threshold:
            try:
                # self.log.error("%s: TangoDCMotor new position  , %s", self.name(), value)
                self.emit("valueChanged", (value,))
                self.old_value = value
            except Exception:
                self.log.error("%s: TangoDCMotor not responding, %s", self.id, "")
                self.old_value = value

    def is_ready(self):
        return self.stateValue == "STANDBY"

    def connect_notify(self, signal):
        if signal == "hardware_object_name,stateChanged":
            self.motorStateChanged(TangoDCMotor.stateDict[self.stateValue])
        elif signal == "limitsChanged":
            self.motorLimitsChanged()
        elif signal == "valueChanged":
            self.motor_positions_changed(self.positionValue)
        self.set_is_ready(True)

    def motorState(self):
        return TangoDCMotor.stateDict[self.stateValue]

    def motorStateChanged(self, state):
        self.stateValue = str(state)
        self.set_is_ready(True)
        logging.info("motor state changed. it is %s " % self.stateValue)
        self.emit("stateChanged", (TangoDCMotor.stateDict[self.stateValue],))

    def get_state(self):
        return TangoDCMotor.stateDict[self.stateValue]

    def get_limits(self):
        try:
            self.log.info(
                "TangoDCMotor.get_limits: trying to get limits for motor_name %s "
                % (self.motor_name)
            )
            limits = self.ho.getMotorLimits(
                self.motor_name
            )  # limitsCommand() # self.ho.getMotorLimits(self.motor_name)
            self.log.info(
                "TangoDCMotor.get_limits: Getting limits for %s -- %s "
                % (self.motor_name, str(limits))
            )
            if numpy.inf in limits:
                limits = numpy.array([-10000, 10000])
        except Exception:
            if self.motor_name in [
                "detector_distance",
                "detector_horizontal",
                "detector_vertical",
            ]:
                info = self.positionChan.getInfo()
                limits = [float(info.min_value), float(info.max_value)]
            # if self.motor_name == 'detector_ts':
            # limits = [96, 1100]
            # elif self.motor_name == 'detector_tx':
            # limits =
            elif self.motor_name == "exposure":
                limits = [float(self.min_value), float(self.max_value)]

        if limits is None:
            try:
                limits = self.get_property("min"), self.get_property("max")
                self.log.info("TangoDCMotor.get_limits: %.4f ***** %.4f" % limits)
                limits = numpy.array(limits)
            except Exception:
                # self.log.info("TangoDCMotor.get_limits: Cannot get limits for %s" % self.name())
                limits = None
        return limits

    def motorLimitsChanged(self):
        self.emit("limitsChanged", (self.get_limits(),))

    def motorIsMoving(self):
        return self.stateValue == "RUNNING" or self.stateValue == "MOVING"

    def motorMoveDone(self, channelValue):
        if self.stateValue == "STANDBY":
            self.emit("moveDone", (self.tangoname, "tango"))

    def motor_positions_changed(self, absolutePosition):
        self.emit("valueChanged", (absolutePosition,))

    def syncQuestionAnswer(self, specSteps, controllerSteps):
        return (
            "0"  # This is only for spec motors. 0 means do not change anything on sync
        )

    def get_value(self):
        return self.positionChan.get_value()

    def convertValue(self, value):
        logging.info("TangoDCMotor: converting value to %s " % str(self.dataType))
        retvalue = value
        if self.dataType in ["short", "int", "long"]:
            retvalue = int(value)
        return retvalue

    def get_motor_mnemonic(self):
        return self.name

    def _set_value(self, value):
        """Move the motor to the required position

        Arguments:
        absolutePosition -- position to move to
        """
        logging.getLogger("TangoClient").info(
            "TangoDCMotor move (%s). Trying to go to %s: type '%s'",
            self.motor_name,
            value,
            type(value),
        )
        value = float(value)
        if not isinstance(value, float) and not isinstance(value, int):
            logging.getLogger("TangoClient").error(
                "Cannot move %s: position '%s' is not a number. It is a %s",
                self.tangoname,
                value,
                type(value),
            )
        logging.info("TangoDCMotor: move. motor will go to %s " % str(value))
        self.log.info("TangoDCMotor.move to absolute position: %.3f" % value)
        logging.getLogger("TangoClient").info(
            "TangoDCMotor move. Trying to go to %s: that is a '%s'",
            value,
            type(value),
        )
        # if abs(self.get_value() - value) > epsilon:
        #     logging.info(
        #         "TangoDCMotor: difference larger then epsilon (%s), executing the move "
        #         % str(epsilon)
        #     )
        self.positionChan.set_value(self.convertValue(value))
        # else:
        #     logging.info(
        #         "TangoDCMotor: not moving really as epsilon is large %s " % str(epsilon)
        #     )
        #     logging.info("TangoDCMotor: self.get_value() %s " % str(self.get_value()))
        #     logging.info("TangoDCMotor: value %s " % str(value))

    def stop(self):
        self.log.info("TangoDCMotor.stop")
        stopcmd = self.get_command_object("Stop")()
        if not stopcmd:
            stopcmd = TangoCommand("stopcmd", "Stop", self.tangoname)
        stopcmd()

    def isSpecConnected(self):
        logging.getLogger().debug("%s: TangoDCMotor.isSpecConnected()" % self.name())
        return (Truehardware_object_name,)


def test():
    hwr = HWR.get_hardware_repository()
    hwr.connect()

    motor = hwr.get_hardware_object("/phi")
    print(motor.get_value())


if __name__ == "__main__":
    test()



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
GenericDiffractometer
"""

import ast
import copy
import enum
import json
import logging
import math
import os
import time
from typing import (
    Dict,
    List,
    Tuple,
    Union,
)

import gevent
import gevent.event
import numpy
from gevent.lock import Semaphore
from pydantic import (
    BaseModel,
    Field,
    ValidationError,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects import sample_centring
from mxcubecore.model import queue_model_objects

__credits__ = ["MXCuBE collaboration"]

__version__ = "2.2."
__status__ = "Draft"


class DiffractometerState:
    """
    Enumeration of diffractometer states
    """

    Created = 0
    Initializing = 1
    On = 2
    Off = 3
    Closed = 4
    Open = 5
    Ready = 6
    Busy = 7
    Moving = 8
    Standby = 9
    Running = 10
    Started = 11
    Stopped = 12
    Paused = 13
    Remote = 14
    Reset = 15
    Closing = 16
    Disable = 17
    Waiting = 18
    Positioned = 19
    Starting = 20
    Loading = 21
    Unknown = 22
    Alarm = 23
    Fault = 24
    Invalid = 25
    Offline = 26

    STATE_DESC = {
        Created: "Created",
        Initializing: "Initializing",
        On: "On",
        Off: "Off",
        Closed: "Closed",
        Open: "Open",
        Ready: "Ready",
        Busy: "Busy",
        Moving: "Moving",
        Standby: "Standby",
        Running: "Running",
        Started: "Started",
        Stopped: "Stopped",
        Paused: "Paused",
        Remote: "Remote",
        Reset: "Reset",
        Closing: "Closing",
        Disable: "Disable",
        Waiting: "Waiting",
        Positioned: " Positioned",
        Starting: "Starting",
        Loading: "Loading",
        Unknown: "Unknown",
        Alarm: "Alarm",
        Fault: "Fault",
        Invalid: "Invalid",
        Offline: "Offline",
    }

    @staticmethod
    def tostring(state):
        return DiffractometerState.STATE_DESC.get(state, "Unknown")


class PhaseEnum(str, enum.Enum):
    centring = "Centring"
    data_collection = "DataCollection"
    beam_location = "BeamLocation"
    transfer = "Transfer"
    unknown = "Unknown"


class PhaseModel(BaseModel):
    value: PhaseEnum = PhaseEnum.unknown


class HeadTypeEnum(str, enum.Enum):
    no_kappa = "NO_KAPPA"
    mini_kappa = "MINI_KAPPA"
    chip = "CHIP"
    plate = "PLATE"


class HolderTypeEnum(str, enum.Enum):
    known_geometry = "KNOWN_GEOMETRY"
    free_geometry = "FREE_GEOMETRY"


class BlockShapeEnum(str, enum.Enum):
    rectangular = "RECTANGULAR"
    elliptical = "ELLIPTICAL"


class CalibrationData(BaseModel):
    top_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top left corner motor position"
    )
    top_right: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Top right corner motor position"
    )
    bottom_left: Tuple[float, float, float] = Field(
        [0, 0, 0], description="Bottom left corner motor position"
    )


class SampleHolderSectionModel(BaseModel):
    calibration_data: CalibrationData | None = None
    section_offset: Tuple[int, int] = Field(
        [0, 0], description="Block offset in grid layout system coordinates x, y"
    )
    block_size: Tuple[float, float] = Field(
        [15, 15], description="Block size horizontal, vertical in mm"
    )
    block_spacing: Tuple[float, float] = Field(
        [15, 15], description="Spacing between blocks horizontal, vertical in mm"
    )
    block_shape: BlockShapeEnum = BlockShapeEnum.rectangular
    number_of_rows: int = Field(6, description="Numer of rows")
    number_of_collumns: int = Field(6, description="Numer of collumns")
    row_labels: List[str] = Field([], description="Row lables")
    column_lables: List[str] = Field([], description="Collumn lables")
    targets_per_block: Tuple[int, int] = Field(
        [20, 20], description="Targets per block dim1 and dim2"
    )


class ChipLayout(BaseModel):
    head_type: HeadTypeEnum = HeadTypeEnum.chip
    holder_type: HolderTypeEnum = HolderTypeEnum.known_geometry
    holder_brand: str = Field("", description="Brand/make of sample holder")
    holder_size: Tuple[float, float] = Field(
        [0, 0], description="Size of sample holder in mm horizontal and vertical"
    )
    sections: List[SampleHolderSectionModel] = []
    calibration_data: CalibrationData | None = None


class GonioHeadConfiguration(BaseModel):
    current: str = Field("", description="Selected chip layout")
    available: Dict[str, ChipLayout] = Field({}, description="Available chip layouts")


class GenericDiffractometer(HardwareObject):
    """
    Abstract base class for diffractometers
    """

    CENTRING_MOTORS_NAME = [
        "phi",
        "phiz",
        "phiy",
        "sampx",
        "sampy",
        "kappa",
        "kappa_phi",
        "beam_x",
        "beam_y",
    ]

    STATE_CHANGED_EVENT = "stateChanged"
    STATUS_CHANGED_EVENT = "statusChanged"
    MOTOR_POSITION_CHANGED_EVENT = "motorPositionsChanged"
    MOTOR_STATUS_CHANGED_EVENT = "motorStatusChanged"

    HEAD_TYPE_MINIKAPPA = "MiniKappa"
    HEAD_TYPE_SMARTMAGNET = "SmartMagnet"
    HEAD_TYPE_PLATE = "Plate"
    HEAD_TYPE_PERMANENT = "Permanent"

    CENTRING_METHOD_MANUAL = "Manual 3-click"
    CENTRING_METHOD_AUTO = "Computer automatic"
    CENTRING_METHOD_MOVE_TO_BEAM = "Move to beam"

    # TODO NBNB FIX THIS CONFUSION!!!
    MANUAL3CLICK_MODE = CENTRING_METHOD_MANUAL
    C3D_MODE = CENTRING_METHOD_AUTO

    PHASE_TRANSFER = "Transfer"
    PHASE_CENTRING = "Centring"
    PHASE_COLLECTION = "DataCollection"
    PHASE_BEAM = "BeamLocation"
    PHASE_UNKNOWN = "Unknown"

    def __init__(self, name):
        HardwareObject.__init__(self, name)

        # Hardware objects ----------------------------------------------------
        self.motor_hwobj_dict = {}
        self.centring_motors_list = None
        self.front_light_switch = None
        self.back_light_switch = None
        self.aperture = None
        self.beamstop = None
        self.cryo = None
        self.capillary = None
        self.use_sc = False

        # Channels and commands -----------------------------------------------
        self.channel_dict = {}
        self.used_channels_list = []
        self.command_dict = {}
        self.used_commands_list = []

        # flag for using sample_centring hwobj or not
        self.use_sample_centring = None

        # Preventing user multiple clicks during manual centring step
        self.waiting_for_click = None  # None = legacy/no-wait mode, True = waiting, False = manual centring in progress
        self.click_lock = Semaphore()

        # time to delay for state polling for controllers
        # not updating state immediately after cmd started
        self.delay_state_polling = None

        self.delay_state_polling = (
            None  # time to delay for state polling for controllers
        )
        # not updating state immediately after cmd started

        # Internal values -----------------------------------------------------
        self.ready_event = None
        self.head_type = GenericDiffractometer.HEAD_TYPE_MINIKAPPA
        self.phase_list = []
        self.grid_direction = None
        self.reversing_rotation = None

        self.beam_position = None
        self.zoom_centre = None
        self.pixels_per_mm_x = None
        self.pixels_per_mm_y = None

        self.current_state = None
        self.current_phase = None
        self.transfer_mode = None
        self.current_centring_procedure = None
        self.current_centring_method = None
        self.current_motor_positions = {}
        self.current_motor_states = {}

        self.fast_shutter_is_open = None
        self.centring_status = {"valid": False}
        self.centring_time = 0
        self.user_confirms_centring = None
        self.user_clicked_event = None
        self.automatic_centring_try_count = 1
        self.omega_reference_par = None
        self.move_to_motors_positions_task = None
        self.move_to_motors_positions_procedure = None

        self.centring_methods = {
            GenericDiffractometer.CENTRING_METHOD_MANUAL: self.start_manual_centring,
            GenericDiffractometer.CENTRING_METHOD_AUTO: self.start_automatic_centring,
            GenericDiffractometer.CENTRING_METHOD_MOVE_TO_BEAM: self.start_move_to_beam,
        }

        self.connect(self, "equipmentReady", self.equipment_ready)
        self.connect(self, "equipmentNotReady", self.equipment_not_ready)

        # HACK
        self.get_motor_positions = self.get_positions

    def init(self):
        super().init()
        # Internal values -----------------------------------------------------
        self.ready_event = gevent.event.Event()
        self.user_clicked_event = gevent.event.AsyncResult()
        self.user_confirms_centring = True

        self.beamstop = self.get_object_by_role("beamstop")
        self.aperture = self.get_object_by_role("aperture")
        self.capillary = self.get_object_by_role("capillary")
        self.cryo = self.get_object_by_role("cryo")

        # Hardware objects ----------------------------------------------------
        # if HWR.beamline.sample_view.camera is not None:
        #     self.image_height = HWR.beamline.sample_view.camera.get_height()
        #     self.image_width = HWR.beamline.sample_view.camera.get_width()
        # else:
        #     self.log.debug(
        #         "Diffractometer: " + "Camera hwobj is not defined"
        #     )

        if HWR.beamline.beam is not None:
            self.beam_position = HWR.beamline.beam.get_beam_position_on_screen()
            self.connect(
                HWR.beamline.beam, "beamPosChanged", self.beam_position_changed
            )
        else:
            self.beam_position = [0, 0]
            self.log.warning("Diffractometer: " + "BeamInfo hwobj is not defined")

        self.front_light_switch = self.get_object_by_role("frontlightswitch")
        self.back_light_switch = self.get_object_by_role("backlightswitch")

        # Channels -----------------------------------------------------------
        self.used_channels_list = self.get_property("used_channels", [])
        if isinstance(self.used_channels_list, str):
            self.used_channels_list = ast.literal_eval(self.used_channels_list)

        for channel_name in self.used_channels_list:
            self.channel_dict[channel_name] = self.get_channel_object(channel_name)
            if self.channel_dict[channel_name] is None:
                continue
            if channel_name == "TransferMode":
                self.connect(
                    self.channel_dict["TransferMode"],
                    "update",
                    self.transfer_mode_changed,
                )
            elif channel_name == "CurrentPhase":
                self.connect(
                    self.channel_dict["CurrentPhase"],
                    "update",
                    self.current_phase_changed,
                )
            elif channel_name == "HeadType":
                self.connect(
                    self.channel_dict["HeadType"], "update", self.head_type_changed
                )
            elif channel_name == "State":
                self.connect(self.channel_dict["State"], "update", self.state_changed)

        # Commands -----------------------------------------------------------
        self.used_commands_list = self.get_property("used_commands", [])
        if isinstance(self.used_commands_list, str):
            self.used_commands_list = ast.literal_eval(self.used_commands_list)

        for command_name in self.used_commands_list:
            self.command_dict[command_name] = self.get_command_object(command_name)

        # Centring motors ----------------------------------------------------
        self.centring_motors_list = self.get_property(
            "centring_motors", GenericDiffractometer.CENTRING_MOTORS_NAME
        )
        if isinstance(self.centring_motors_list, str):
            self.centring_motors_list = ast.literal_eval(self.centring_motors_list)

        queue_model_objects.CentredPosition.set_diffractometer_motor_names(
            *self.centring_motors_list
        )

        for motor_name in self.centring_motors_list:
            # NBNB TODO refactor configuration, and set properties directly (see below)
            temp_motor_hwobj = self.get_object_by_role(motor_name)
            if temp_motor_hwobj is not None:
                self.motor_hwobj_dict[motor_name] = temp_motor_hwobj
                self.connect(temp_motor_hwobj, "stateChanged", self.motor_state_changed)
                self.connect(
                    temp_motor_hwobj, "valueChanged", self.centring_motor_moved
                )

                if motor_name == "phi":
                    self.connect(
                        temp_motor_hwobj,
                        "valueChanged",
                        self.emit_diffractometer_moved,
                    )
                    #
                    # A work-around to make mesh grid 'Rotate to' feature work.
                    #
                    # MXCuBE requires that diffractometer HWO have a `self.centringPhi`
                    # attribute, used to change 'omega' rotation when running
                    # mesh 'Rotate to' routine.
                    #
                    # See https://github.com/mxcube/mxcubecore/issues/1360 for details.
                    #
                    self.centringPhi = self.motor_hwobj_dict["phi"]
                elif motor_name == "zoom":
                    self.connect(
                        temp_motor_hwobj,
                        "predefinedPositionChanged",
                        self.zoom_motor_predefined_position_changed,
                    )
                    self.connect(
                        temp_motor_hwobj, "stateChanged", self.zoom_motor_state_changed
                    )
            else:
                self.log.warning(
                    "Diffractometer: Motor "
                    + "%s listed in the centring motor list, but not initalized"
                    % motor_name
                )

        # sample changer -----------------------------------------------------
        if HWR.beamline.sample_changer is None:
            self.log.warning("Diffractometer: Sample Changer is not defined")
        else:
            # By default use sample changer if it's defined and transfer_mode
            # is set to SAMPLE_CHANGER.
            # if not defined, set use_sc to True
            if self.transfer_mode is None or self.transfer_mode == "SAMPLE_CHANGER":
                self.use_sc = True

        try:
            self.use_sample_centring = self.get_property("sample_centring")
            if self.use_sample_centring:
                self.centring_phi = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phi"], direction=-1
                )
                self.centring_phiz = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phiz"]
                )
                self.centring_phiy = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["phiy"], direction=-1
                )
                self.centring_sampx = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["sampx"]
                )
                self.centring_sampy = sample_centring.CentringMotor(
                    self.motor_hwobj_dict["sampy"]
                )
        except Exception:
            self.log.exception("")

        try:
            self.delay_state_polling = self.get_property("delay_state_polling")
        except Exception:
            self.log.exception("")

        # Other parameters ---------------------------------------------------
        self.zoom_centre = self.get_property("zoom_centre")
        if self.zoom_centre is None:
            self.zoom_centre = {"x": 0, "y": 0}
            self.log.warning(
                "Diffractometer: zoom centre not configured. Using default."
            )
        elif isinstance(self.zoom_centre, str):
            self.zoom_centre = ast.literal_eval(self.zoom_centre)
        elif isinstance(self.zoom_centre, dict):
            self.zoom_centre = self.zoom_centre
        else:
            _m = "Diffractometer: zoom centre misconfigured. not a dict"
            raise ValueError(_m)

        if not all(key in self.zoom_centre for key in ["x", "y"]):
            _m = "Diffractometer: zoom centre misconfigured. missing x or y"
            raise ValueError(_m)

        self.reversing_rotation = self.get_property("reversing_rotation")
        # grid_direction describes how a grid is collected
        # 'fast' is collection direction and 'slow' describes
        # move to the next collection line
        self.grid_direction = self.get_property("grid_direction")
        if self.grid_direction is None:
            self.grid_direction = {"fast": (0, 1), "slow": (1, 0), "omega_ref": 0}
            self.log.warning(
                "Diffractometer: Grid direction not configured. Using default."
            )
        elif isinstance(self.grid_direction, str):
            self.grid_direction = ast.literal_eval(self.grid_direction)
        elif not isinstance(self.grid_direction, dict):
            _m = "Diffractometer: grid direction misconfigured. not a dict"
            raise ValueError(_m)

        if not all(key in self.grid_direction for key in ["fast", "slow", "omega_ref"]):
            _m = "Diffractometer: grid direction misconfigured. missing fast, slow, or omega_ref"
            raise ValueError(_m)

        self.phase_list = self.get_property(
            "phase_list",
            [
                GenericDiffractometer.PHASE_TRANSFER,
                GenericDiffractometer.PHASE_CENTRING,
                GenericDiffractometer.PHASE_COLLECTION,
                GenericDiffractometer.PHASE_BEAM,
            ],
        )
        if isinstance(self.phase_list, str):
            self.phase_list = ast.literal_eval(self.phase_list)

    # to make it compatible
    def __getattr__(self, attr):
        if attr.startswith("__"):
            raise AttributeError(attr)
        else:
            if attr == "currentCentringProcedure":
                return self.current_centring_procedure
            if attr == "centringStatus":
                return self.centring_status
            if attr == "imageClicked":
                return self.image_clicked
            if attr == "cancelCentringMethod":
                return self.cancel_centring_method
            if attr == "pixelsPerMmY":
                return self.pixels_per_mm_x
            if attr == "pixelsPerMmZ":
                return self.pixels_per_mm_y
            return HardwareObject.__getattr__(self, attr)

    # Contained Objects
    # NBNB Temporary hack - should be cleaned up together with configuration
    @property
    def omega(self):
        """omega motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phi")

    @property
    def kappa(self):
        """kappa motor object

        Returns:
            AbstractActuator
        """
        return self.get_object_by_role("kappa")

    @property
    def kappa_phi(self):
        """kappa motor object

        Returns:
            AbstractActuator
        """
        return self.get_object_by_role("kappa_phi")

    @property
    def centring_x(self):
        """centring_x motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("sampx")

    @property
    def centring_y(self):
        """centring_y motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("sampy")

    @property
    def alignment_x(self):
        """alignment_x motor object (also used as graphics.focus)

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("focus")

    @property
    def alignment_y(self):
        """alignment_y motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phiy")

    @property
    def alignment_z(self):
        """alignment_z motor object

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("phiz")

    @property
    def zoom(self):
        """zoom motor object

        NBNB TBD This is supposedly now living in AbstractSampleView,
        And the property was removed in an earlier PR (20251021, 15:24 Elmir Jagudin)
        BUT 1) there does not seem to be any mechanism to set it from config there
        2) it is configured on the diffractometer according to available config files
        3) accessing diffractometer.zoom is all through the code
        This needs global refactoring, but meanwhile this property should remain

        Returns:
            AbstractActuator
        """
        return self.motor_hwobj_dict.get("zoom")

    def is_ready(self):
        """
        Detects if device is ready
        """
        return self.current_state == DiffractometerState.tostring(
            DiffractometerState.Ready
        )

    def wait_device_not_ready(self, timeout=5):
        with gevent.Timeout(timeout, Exception("Timeout waiting for device not ready")):
            while self.is_ready():
                time.sleep(0.01)

    def wait_device_ready(self, timeout=30):
        """Waits when diffractometer status is ready:

        :param timeout: timeout in second
        :type timeout: int
        """
        with gevent.Timeout(timeout, Exception("Timeout waiting for device ready")):
            while not self.is_ready():
                time.sleep(0.01)

    wait_ready = wait_device_ready

    def execute_server_task(self, method, timeout=30, *args):
        """Method is used to execute commands and wait till
           diffractometer is in ready state

        :param method: method to be executed
        :type method: instance
        :param timeout: timeout in seconds
        :type timeout: seconds
        """
        # self.ready_event.clear()
        self.current_state = DiffractometerState.tostring(DiffractometerState.Busy)
        method(*args)
        time.sleep(5)
        # gevent.sleep(2)
        self.wait_device_ready(timeout)
        self.ready_event.set()

    def in_plate_mode(self):
        """Returns True if diffractometer in plate mod

        :returns: boolean
        """
        return self.head_type == GenericDiffractometer.HEAD_TYPE_PLATE

    def get_head_type(self):
        """Returns head type

        :returns: string
        """
        return self.head_type

    def use_sample_changer(self):
        """Returns True if sample changer is in use

        :returns: boolean
        """
        return self.use_sc

    def set_use_sc(self, flag):
        """Sets use_sc flag, that indicates if sample changer is used

        :param flag: use sample changer flag
        :type flag: boolean
        """
        if flag:
            # check both transfer_mode and sample_Changer
            if HWR.beamline.sample_changer is None:
                self.log.error("Diffractometer: Sample " + "Changer is not available")
                return False

            if (
                self.transfer_mode is None
                or self.channel_dict["TransferMode"].get_value() == "SAMPLE_CHANGER"
            ):
                # if transferMode is not defined, ignore the checkup
                self.use_sc = True
            else:
                self.log.error(
                    "Diffractometer: Set the "
                    + "diffractometer TransferMode to SAMPLE_CHANGER first!!"
                )
                return False
        else:
            self.use_sc = False
        return True

    def transfer_mode_changed(self, transfer_mode):
        """
        Descript. :
        """
        self.log.info("current_transfer_mode is set to %s" % transfer_mode)
        self.transfer_mode = transfer_mode
        if transfer_mode != "SAMPLE_CHANGER":
            self.use_sc = False
        self.emit("minidiffTransferModeChanged", (transfer_mode,))

    def get_transfer_mode(self):
        """ """
        return self.transfer_mode

    def get_current_phase(self):
        """
        Descript. :
        """
        return self.current_phase

    def get_grid_direction(self):
        """
        Descript. :
        """
        return self.grid_direction

    def get_available_centring_methods(self):
        """
        Descript. :
        """
        return self.centring_methods.keys()

    def get_current_centring_method(self):
        """
        Descript. :
        """
        return self.current_centring_method

    def is_reversing_rotation(self):
        """ """
        return self.reversing_rotation is True

    def beam_position_changed(self, value):
        """
        Descript. :
        """
        self.beam_position = list(value)

    # def get_motor_positions(self):
    #    return

    # TODO rename to get_motor_positions
    def get_positions(self):
        """
        Descript. :
        """

        self.current_motor_positions["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        self.current_motor_positions["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x

        return self.current_motor_positions

    def get_motors(self):
        """Get motor_name:Motor dictionary"""
        return self.motor_hwobj_dict.copy()

    def get_snapshot(self):
        """
        Get snapshot from sample view

        Returns:
            bytes: A bytes object of the current camera image.
        """
        if HWR.beamline.sample_view:
            return HWR.beamline.sample_view.get_snapshot()

    def save_snapshot(self, filename):
        if HWR.beamline.sample_view:
            return HWR.beamline.sample_view.save_snapshot(filename)

    def get_pixels_per_mm(self):
        """
        Returns tuple with pixels_per_mm_x and pixels_per_mm_y

        :returns: list with two floats
        """
        return (self.pixels_per_mm_x, self.pixels_per_mm_y)

    def get_phase_list(self):
        """
        Returns list of available phases

        :returns: list with str
        """
        return self.phase_list

    def start_centring_method(self, method, sample_info=None, wait=False):
        """ """

        if self.current_centring_method is not None:
            self.log.error(
                "Diffractometer: already in centring method %s"
                % self.current_centring_method
            )
            return
        curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
        self.centring_status = {
            "valid": False,
            "startTime": curr_time,
            "angleLimit": None,
        }
        self.emit_centring_started(method)

        try:
            centring_method = self.centring_methods[method]
        except KeyError as diag:
            self.log.error("Diffractometer: unknown centring method (%s)" % str(diag))
            self.emit_centring_failed()
        else:
            try:
                centring_method(sample_info, wait_result=wait)
            except Exception:
                self.log.exception("Diffractometer: problem while centring")
                self.emit_centring_failed()

    def cancel_centring_method(self, reject=False):
        if self.current_centring_procedure is not None:
            try:
                self.current_centring_procedure.kill()
            except Exception:
                self.log.exception(
                    "Diffractometer: problem aborting the centring method"
                )
            try:
                # TODO... do we need this at all?
                # fun = self.cancel_centring_methods[self.current_centring_method]
                pass
            except KeyError:
                self.emit_centring_failed()
            else:
                try:
                    fun()
                except Exception:
                    self.emit_centring_failed()
        else:
            self.emit_centring_failed()
        self.emit_progress_message("")
        if reject:
            self.reject_centring()

    def start_manual_centring(self, sample_info=None, wait_result=None):
        self.emit_progress_message(f"{self.CENTRING_METHOD_MANUAL} centring...")
        if self.use_sample_centring:
            self.current_centring_procedure = sample_centring.start(
                {
                    "phi": self.centring_phi,
                    "phiy": self.centring_phiy,
                    "sampx": self.centring_sampx,
                    "sampy": self.centring_sampy,
                    "phiz": self.centring_phiz,
                },
                self.pixels_per_mm_x,
                self.pixels_per_mm_y,
                self.beam_position[0],
                self.beam_position[1],
            )
        else:
            self.current_centring_procedure = gevent.spawn(self.manual_centring)
        self.current_centring_procedure.link(self.centring_done)

    def start_automatic_centring(
        self, sample_info=None, loop_only=False, wait_result=None
    ):
        self.emit_progress_message("Automatic centring...")

        while self.automatic_centring_try_count > 0:
            if self.use_sample_centring:
                self.current_centring_procedure = sample_centring.start_auto(
                    HWR.beamline.sample_view,
                    {
                        "phi": self.centring_phi,
                        "phiy": self.centring_phiy,
                        "sampx": self.centring_sampx,
                        "sampy": self.centring_sampy,
                        "phiz": self.centring_phiz,
                    },
                    self.pixels_per_mm_x,
                    self.pixels_per_mm_y,
                    self.beam_position[0],
                    self.beam_position[1],
                    msg_cb=self.emit_progress_message,
                    new_point_cb=lambda point: self.emit(
                        "newAutomaticCentringPoint", (point,)
                    ),
                )
            else:
                self.current_centring_procedure = gevent.spawn(self.automatic_centring)
            self.current_centring_procedure.link(self.centring_done)

            if wait_result:
                self.ready_event.wait()
                self.ready_event.clear()

            self.automatic_centring_try_count -= 1
        self.automatic_centring_try_count = 1

    def start_move_to_beam(
        self, coord_x=None, coord_y=None, omega=None, wait_result=None
    ):
        """
        Descript. :
        """
        try:
            self.emit_progress_message("Move to beam...")
            self.centring_time = time.time()
            curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
            self.centring_status = {
                "valid": True,
                "startTime": curr_time,
                "endTime": curr_time,
            }
            if coord_x is None and coord_y is None:
                coord_x = self.beam_position[0]
                coord_y = self.beam_position[1]

            motors = self.get_centred_point_from_coord(
                coord_x, coord_y, return_by_names=True
            )
            if omega is not None:
                motors["phi"] = omega

            self.centring_status["motors"] = motors
            self.centring_status["valid"] = True
            self.centring_status["angleLimit"] = True
            self.emit_progress_message("")
            self.accept_centring()
            self.current_centring_method = None
            self.current_centring_procedure = None
        except Exception:
            logging.exception("Diffractometer: Could not complete 2D centring")

    def centring_done(self, centring_procedure):
        try:
            motor_pos = centring_procedure.get()
            if isinstance(motor_pos, gevent.GreenletExit):
                raise motor_pos
        except Exception:
            logging.exception("Could not complete centring")
            self.emit_centring_failed()
        else:
            self.emit_progress_message("Moving sample to centred position...")
            self.emit_centring_moving()

            try:
                self.log.debug(
                    "Centring finished. Moving motors to position %s" % str(motor_pos)
                )
                self.move_to_motors_positions(motor_pos, wait=True)
            except Exception:
                logging.exception("Could not move to centred position")
                self.emit_centring_failed()
            else:
                # if 3 click centring move -180. well. dont, in principle the calculated
                # centred positions include omega to initial position
                pass
                # if not self.in_plate_mode():
                #    self.log.debug("Centring finished. Moving omega back to initial position")
                #    self.motor_hwobj_dict['phi'].set_value_relative(-180, timeout=None)
                #    self.log.debug("         Moving omega done")

            if (
                self.current_centring_method
                == GenericDiffractometer.CENTRING_METHOD_AUTO
            ):
                self.emit("newAutomaticCentringPoint", motor_pos)
            self.ready_event.set()
            self.centring_time = time.time()
            self.emit_centring_successful()
            self.emit_progress_message("")

    def manual_centring(self):
        raise NotImplementedError

    def automatic_centring(self):
        raise NotImplementedError

    def centring_motor_moved(self, pos):
        if time.time() - self.centring_time > 1.0:
            self.invalidate_centring()
        self.emit_diffractometer_moved()

    def invalidate_centring(self):
        if self.current_centring_procedure is None and self.centring_status["valid"]:
            self.centring_status = {"valid": False}
            self.emit_progress_message("")
            self.emit("centringInvalid", ())

    def emit_diffractometer_moved(self, *args):
        self.emit("diffractometerMoved", ())

    def motor_positions_to_screen(self, centred_positions_dict):
        if self.use_sample_centring:
            self.update_zoom_calibration()
            if None in (self.pixels_per_mm_x, self.pixels_per_mm_y):
                return 0, 0
            phi_angle = math.radians(
                self.centring_phi.direction * self.centring_phi.get_value()
            )
            sampx = self.centring_sampx.direction * (
                centred_positions_dict["sampx"] - self.centring_sampx.get_value()
            )
            sampy = self.centring_sampy.direction * (
                centred_positions_dict["sampy"] - self.centring_sampy.get_value()
            )
            phiy = self.centring_phiy.direction * (
                centred_positions_dict["phiy"] - self.centring_phiy.get_value()
            )
            phiz = self.centring_phiz.direction * (
                centred_positions_dict["phiz"] - self.centring_phiz.get_value()
            )
            rot_matrix = numpy.matrix(
                [
                    math.cos(phi_angle),
                    -math.sin(phi_angle),
                    math.sin(phi_angle),
                    math.cos(phi_angle),
                ]
            )
            rot_matrix.shape = (2, 2)
            inv_rot_matrix = numpy.array(rot_matrix.I)
            dx, dy = (
                numpy.dot(numpy.array([sampx, sampy]), inv_rot_matrix)
                * self.pixels_per_mm_x
            )

            x = (phiy * self.pixels_per_mm_x) + self.beam_position[0]
            y = dy + (phiz * self.pixels_per_mm_y) + self.beam_position[1]

            return x, y
        else:
            raise NotImplementedError

    def move_to_centred_position(self, centred_position):
        self.move_motors(centred_position)

    def move_to_motors_positions(self, motors_positions, wait=False):
        self.emit_progress_message("Moving to motors positions...")
        self.move_to_motors_positions_procedure = gevent.spawn(
            self.move_motors, motors_positions
        )
        self.move_to_motors_positions_procedure.link(self.move_motors_done)
        if wait:
            self.wait_device_ready(10)

    def move_motors(self, motor_positions, timeout=15):
        """
        Moves diffractometer motors to the requested positions

        :param motors_dict: dictionary with motor names or hwobj
                            and target values.
        :type motors_dict: dict
        """
        if not isinstance(motor_positions, dict):
            motor_positions = motor_positions.as_dict()

        self.wait_device_ready(timeout)

        for motor in motor_positions.keys():
            position = motor_positions[motor]
            self.log.debug(f"moving motor {motor} to position {position}")
            if isinstance(motor, str):
                motor_role = motor
                motor = self.motor_hwobj_dict.get(motor_role)
                if None in (motor, position):
                    continue
            motor.set_value(position)
        self.wait_device_ready(timeout)

        if self.delay_state_polling is not None and self.delay_state_polling > 0:
            # delay polling for state in the
            # case of controller not reporting MOVING immediately after cmd
            gevent.sleep(self.delay_state_polling)

        self.wait_device_ready(timeout)

    def move_motors_done(self, move_motors_procedure):
        """
        Descript. :
        """
        self.move_to_motors_positions_procedure = None
        self.emit_progress_message("")

    def move_to_beam(self, x, y, omega=None):
        """
        Descript. : function to create a centring point based on all motors
                    positions.
        """
        try:
            pos = self.get_centred_point_from_coord(x, y, return_by_names=False)
            if omega is not None:
                pos["phiMotor"] = omega
            self.move_to_motors_positions(pos)
        except Exception:
            self.log.exception("Diffractometer: could not center to beam, aborting")

    def image_clicked(
        self, x: float, y: float, xi: float | None = None, yi: float | None = None
    ):
        """Handles a user click sent from the frontend during the manual centring.

        This method is called by the backend when the user clicks on the sample
        image in the frontend.

        The attribute `self.waiting_for_click` controls whether the click should
        be accepted or ignored:
          - None: click is accepted (legacy)
          - True: waiting for a click, accept it and mark it as received
          - False: already received a click, ignore further clicks

        Args:
            x: X coordinate of the click.
            y: Y coordinate of the click.
            xi: ...
            yi: ...

        Raises:
            RuntimeError: If a click is received while a previous one is still being processed.
        """
        with self.click_lock:
            # "waiting for click" logic is not implememted (legacy) or it is actually waiting for a click
            if self.waiting_for_click is None or self.waiting_for_click:
                if self.waiting_for_click:
                    self.waiting_for_click = False
                if self.use_sample_centring:
                    sample_centring.user_click(x, y)
                else:
                    self.user_clicked_event.set((x, y))
            # Already received a click, ignore further clicks
            else:
                self.log.warning(
                    "User attempted to click while the previous centring step was still in progress. Click ignored"
                )
                err_msg = "Click ignored: a centring step is still being processed. Please wait before clicking again."
                raise RuntimeError(err_msg)

    def accept_centring(self):
        """
        Descript. :
        Arg.      " fully_centred_point. True if 3 click centring
                    else False
        """
        self.centring_status["valid"] = True
        self.centring_status["accepted"] = True
        centring_status = self.get_centring_status()
        if "motors" not in centring_status:
            centring_status["motors"] = self.get_positions()
        self.emit("centringAccepted", (True, centring_status))
        self.emit("fsmConditionChanged", "centering_position_accepted", True)

    def reject_centring(self):
        if self.current_centring_procedure:
            self.current_centring_procedure.kill()
        self.centring_status = {"valid": False}
        self.emit_progress_message("")
        self.emit("centringAccepted", (False, self.get_centring_status()))
        self.emit("fsmConditionChanged", "centering_position_accepted", False)

    def emit_centring_started(self, method):
        self.current_centring_method = method
        self.emit("centringStarted", (method, False))

    def emit_centring_moving(self):
        self.emit("centringMoving", ())

    def emit_centring_failed(self):
        self.centring_status = {"valid": False}
        method = self.current_centring_method
        self.current_centring_method = None
        self.current_centring_procedure = None
        self.emit("centringFailed", (method, self.get_centring_status()))

    def emit_centring_successful(self):
        if self.current_centring_procedure is not None:
            curr_time = time.strftime("%Y-%m-%d %H:%M:%S")
            self.centring_status["endTime"] = curr_time

            motor_pos = self.current_centring_procedure.get()
            self.centring_status["motors"] = self.convert_from_obj_to_name(motor_pos)
            self.centring_status["method"] = self.current_centring_method
            self.centring_status["valid"] = True

            method = self.current_centring_method
            self.emit("centringSuccessful", (method, self.get_centring_status()))
            self.current_centring_method = None
            self.current_centring_procedure = None
        else:
            self.log.debug(
                "Diffractometer: Trying to emit "
                + "centringSuccessful outside of a centring"
            )

    def emit_progress_message(self, msg=None):
        self.emit("progressMessage", (msg,))

    def get_centring_status(self):
        return copy.deepcopy(self.centring_status)

    def get_centred_point_from_coord(self, x, y, return_by_names=None):
        raise NotImplementedError

    def get_point_between_two_points(
        self, point_one, point_two, frame_num, frame_total
    ):
        """
        Method returns a centring point between two centring points
        It is used to get a position on a helical line based on
        frame number and total frame number
        """
        new_point = {}
        point_one = point_one.as_dict()
        point_two = point_two.as_dict()
        for motor in point_one.keys():
            new_motor_pos = (
                frame_num
                / float(frame_total)
                * abs(point_one[motor] - point_two[motor])
                + point_one[motor]
            )
            new_motor_pos += 0.5 * (point_two[motor] - point_one[motor]) / frame_total
            new_point[motor] = new_motor_pos
        return new_point

    def convert_from_obj_to_name(self, motor_pos):
        motors = {}
        for motor_role in self.centring_motors_list:
            motor_obj = self.get_object_by_role(motor_role)
            try:
                motors[motor_role] = motor_pos[motor_obj]
            except KeyError:
                if motor_obj:
                    motors[motor_role] = motor_obj.get_value()
        motors["beam_x"] = (
            self.beam_position[0] - self.zoom_centre["x"]
        ) / self.pixels_per_mm_y
        motors["beam_y"] = (
            self.beam_position[1] - self.zoom_centre["y"]
        ) / self.pixels_per_mm_x
        return motors

    def visual_align(self, point_1, point_2):
        """
        Descript. :
        """
        return

    def move_omega_relative(self, relative_angle):
        """
        Descript. :
        """
        return

    def get_osc_limits(self):
        """Returns osc limits"""
        return

    def get_osc_max_speed(self):
        return

    def get_scan_limits(self, speed=None, num_images=None, exp_time=None):
        """
        Gets scan limits. Necessary for example in the plate mode
        where osc range is limited
        """
        return

    def set_phase(self, phase, timeout=None):
        """Sets diffractometer to selected phase
           By default available phase is Centring, BeamLocation,
           DataCollection, Transfer

        :param phase: phase
        :type phase: string
        :param timeout: timeout in sec
        :type timeout: int
        """
        if timeout:
            self.ready_event.clear()
            set_phase_task = gevent.spawn(
                self.execute_server_task,
                self.command_dict["startSetPhase"],
                timeout,
                phase,
            )
            self.ready_event.wait()
            self.ready_event.clear()
        else:
            self.command_dict["startSetPhase"](phase)

    def update_zoom_calibration(self):
        self.pixels_per_mm_x = 1.0 / self.channel_dict["CoaxCamScaleX"].get_value()
        self.pixels_per_mm_y = 1.0 / self.channel_dict["CoaxCamScaleY"].get_value()
        self.emit("pixelsPerMmChanged", ((self.pixels_per_mm_x, self.pixels_per_mm_y)))

    def zoom_motor_state_changed(self, state):
        self.emit("zoomMotorStateChanged", (state,))
        self.emit("minidiffStateChanged", (state,))

    def zoom_motor_predefined_position_changed(self, position_name, offset):
        self.update_zoom_calibration()
        self.emit("zoomMotorPredefinedPositionChanged", (position_name, offset))

    def equipment_ready(self):
        self.emit("minidiffReady", ())

    def equipment_not_ready(self):
        self.emit("minidiffNotReady", ())

    def motor_state_changed(self, state):
        """
        Descript. :
        """
        self.emit("minidiffStateChanged", (state,))

    def current_phase_changed(self, current_phase):
        """
        Descript. :
        """
        self.current_phase = current_phase
        if current_phase != GenericDiffractometer.PHASE_UNKNOWN:
            logging.getLogger("GUI").info(
                "Diffractometer: Current phase changed to %s" % current_phase
            )
        self.emit("minidiffPhaseChanged", (current_phase,))

    def sample_is_loaded_changed(self, sample_is_loaded):
        """
        Descript. :
        """
        self.sample_is_loaded = sample_is_loaded
        self.emit("minidiffSampleIsLoadedChanged", (sample_is_loaded,))

    def head_type_changed(self, head_type):
        """
        Descript. :
        """
        self.head_type = head_type
        self.emit("minidiffHeadTypeChanged", (head_type,))

        if "SampleIsLoaded" not in str(self.used_channels_list):
            return
        try:
            self.disconnect(
                self.channel_dict["SampleIsLoaded"],
                "update",
                self.sample_is_loaded_changed,
            )
        except Exception:
            self.log.exception("")

        if (
            head_type == GenericDiffractometer.HEAD_TYPE_MINIKAPPA
            or head_type == GenericDiffractometer.HEAD_TYPE_SMARTMAGNET
        ):
            self.connect(
                self.channel_dict["SampleIsLoaded"],
                "update",
                self.sample_is_loaded_changed,
            )
        else:
            self.log.info(
                "Diffractometer: SmartMagnet "
                + "is not available, only works for Minikappa and SmartMagnet head"
            )

    def move_kappa_and_phi(self, kappa, kappa_phi):
        return

    def close_kappa(self):
        """
        Descript. :
        """
        return

    def get_osc_dynamic_limits(self):
        return (-10000, 10000)

    def zoom_in(self):
        return

    def zoom_out(self):
        return

    def save_centring_positions(self):
        pass

    def force_emit_signals(self):
        for motor_hwobj in self.motor_hwobj_dict.values():
            motor_hwobj.force_emit_signals()

    def get_head_configuration(self) -> Union[GonioHeadConfiguration, None]:
        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        data = None

        if os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "r") as _f:
                chip_def = json.load(_f)

                try:
                    data = GonioHeadConfiguration(**chip_def)
                except ValidationError:
                    self.log.exception("Validation error in %s" % chip_def_fpath)

        return data

    def set_head_configuration(self, str_data: str) -> None:
        data = json.loads(str_data)

        chip_def_fpath = self.get_property("chip_definition_file", "")
        chip_def_fpath = HWR.get_hardware_repository().find_in_repository(
            chip_def_fpath
        )

        if os.path.isfile(chip_def_fpath):
            with open(chip_def_fpath, "w+") as _f:
                try:
                    GonioHeadConfiguration(**data)
                except ValidationError:
                    self.log.exception("Validation error in %s" % chip_def_fpath)
                else:
                    _f.write(json.dumps(data, indent=4))

    def set_chip_layout(self, layout_name: str) -> bool:
        data = self.get_head_configuration().dict()
        data["current"] = layout_name
        self.set_head_configuration(json.dumps(data))
        return True



from mxcubecore.HardwareObjects.SardanaMotor import SardanaMotor


class AutoPowerSardanaMotor(SardanaMotor):
    """Adds auto power-on feature to a SardanaMotor.

    This hardware object powers on the underlying motor,
    before it is moved.

    It is assumed that the sardana movable have a boolean
    'PowerOn' property. Setting this property to True will
    power on the underlying motors of the movable.
    """

    def __init__(self, name):
        super().__init__(name)
        self.power_on_channel = None

    def init(self):
        super().init()

        self.power_on_channel = self.add_channel(
            {
                "type": "sardana",
                "name": f"{self.actuator_name}PowerOn",
                "taurusname": self.taurusname,
            },
            "PowerOn",
        )

    def _set_value(self, value):
        # power on the motor
        self.power_on_channel.set_value(True)
        # start the motor movement
        super()._set_value(value)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"


# import DeviceProxy function and DevState:

from mxcubecore.BaseHardwareObjects import HardwareObject

last_centred_position = [200, 200]


class Centring(HardwareObject):
    """
    Description:     This class controls the operation of Tango Motor
    """

    def __init__(self, *args):
        """
        Description:
        """
        HardwareObject.__init__(self, *args)
        self.proxyMotor = None  # Hardware object to change motor attributes
        self.motor_name = None  # Tango name of DeviceServer controlling the motor

    def init(self):
        self.gonioAxes = []

        # self.gonioAxes = []
        # for axis in self['gonioAxes']:
        #  self.gonioAxes.append({'type':axis.type,'direction':eval(axis.direction),\
        #                   'motor_name':axis.motorname,'motor_HO':
        # HardwareRepository.get_hardware_repository().get_hardware_object(axis.motorHO)
        # })

        print("Centring Init")
        print("-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+")

    def initCentringProcedure(self):
        print("initCentringProcedure(self)")
        """
        Descript. : call before starting rotate-click sequence
        """
        self.centringDataTensor = []
        self.centringDataMatrix = []
        self.motorConstraints = []

        # self.centeingPoints = [Point(0.0,0.0), Point(0.0,0.0), Point(0.0,0.0)]  # array of 3 points, initial values - zeroes
        # print( "self.centeingPoints[0] = (", self.centeingPoints[0].x, ", ", self.centeingPoints[0].y, ")")
        # print( "self.centeingPoints[1] = (", self.centeingPoints[1].x, ", ", self.centeingPoints[1].y, ")")
        # print( "self.centeingPoints[2] = (", self.centeingPoints[2].x, ", ", self.centeingPoints[2].y, ")")

        print("Centring initCentringProcedure")
        print("-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+")

    def appendCentringDataPoint(self, camera_coordinates):
        """
        Descript. : call after each click and send click points - but relative in mm
        """
        # self.centringDataTensor.append(self.factor_matrix())
        self.centringDataMatrix.append(
            self.camera_coordinates_to_vector(camera_coordinates)
        )

        ll = len(self.centringDataMatrix)
        print(("len(self.centringDataMatrix) = ", ll))

        print(("self.centringDataMatrix = ", self.centringDataMatrix[ll - 1]))

        print("appendCentringDataPoint(self)")
        print("-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+")

    #    def setCentringDataPoint(self, i, camera_coordinates):
    #        print("camera_coordinates['X'] = ",  camera_coordinates['X'])
    #        print("camera_coordinates['Y'] = ",  camera_coordinates['Y'])
    #        #self.centeingPoints[i] =  point(camera_coordinates['X'], camera_coordinates['Y']])
    #        print("appendCentringDataPoint(self)")
    #        print( '-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+')

    def centeredPosition(self, return_by_name=False):
        """
        Descript. : call after appending the last click.
        Return    : {motorHO:position} dictionary.

        M=numpy.zeros(shape=(self.translationAxesCount,self.translationAxesCount))
        V=numpy.zeros(shape=(self.translationAxesCount))

        for l in range(0,self.translationAxesCount):
           for i in range (0,len(self.centringDataMatrix)):
              for k in range(0,len(self.cameraAxes)):
                 V[l] += self.centringDataTensor[i][l][k]*self.centringDataMatrix[i][k]
           for m in range(0,self.translationAxesCount):
              for i in range (0,len(self.centringDataMatrix)):
                 for k in range(0,len(self.cameraAxes)):
                    M[l][m] += self.centringDataTensor[i][l][k]*self.centringDataTensor[i][m][k]
        tau_cntrd = numpy.dot(numpy.linalg.pinv(M,rcond=1e-6),V)

        tau_cntrd = self.apply_constraints(M,tau_cntrd)
        """
        # return self.vector_to_centred_positions( - tau_cntrd +
        # self.translation_datum(), return_by_name)
        print("centeredPosition(self)")
        print("-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+")

        return self.vector_to_centred_positions(123, 456)

    def vector_to_centred_positions(self, vector, return_by_name=False):
        dic = {}
        index = 0
        # for axis in self.gonioAxes:
        #  if axis['type'] == "translation":
        #     if return_by_name:
        #         dic[axis['motor_name']]=vector[index]
        #     else:
        #         dic[axis['motor_HO']]=vector[index]
        #     index += 1

        print("vector_to_centred_positions(self)")
        print("-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+")

        return dic

    def camera_coordinates_to_vector(self, camera_coordinates_dictionary):
        vector = []

        vector.append(camera_coordinates_dictionary["X"])
        vector.append(camera_coordinates_dictionary["Y"])

        return vector
